# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""ExplainLoader."""

import math
import os
import re
import threading
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Dict, Iterable, List, Optional, Union

from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.datavisual.data_access.file_handler import FileHandler
from mindinsight.explainer.common.enums import ExplainFieldsEnum
from mindinsight.explainer.common.log import logger
from mindinsight.explainer.manager.explain_parser import ExplainParser
from mindinsight.utils.exceptions import ParamValueError, UnknownError

_NAN_CONSTANT = 'NaN'
_NUM_DIGITS = 6

_EXPLAIN_FIELD_NAMES = [
    ExplainFieldsEnum.SAMPLE_ID,
    ExplainFieldsEnum.BENCHMARK,
    ExplainFieldsEnum.METADATA,
]

_SAMPLE_FIELD_NAMES = [
    ExplainFieldsEnum.GROUND_TRUTH_LABEL,
    ExplainFieldsEnum.INFERENCE,
    ExplainFieldsEnum.EXPLANATION,
    ExplainFieldsEnum.HIERARCHICAL_OCCLUSION
]


class _LoaderStatus(Enum):
    STOP = 'STOP'
    LOADING = 'LOADING'
    PENDING = 'PENDING'
    LOADED = 'LOADED'


def _round(score):
    """Take round of a number to given precision."""
    try:
        return round(score, _NUM_DIGITS)
    except TypeError:
        return score


class ExplainLoader:
    """ExplainLoader which manage the record in the summary file."""

    def __init__(self,
                 loader_id: str,
                 summary_dir: str):

        self._parser = ExplainParser(summary_dir)

        self._loader_info = {
            'loader_id': loader_id,
            'summary_dir': summary_dir,
            'create_time': os.stat(summary_dir).st_ctime,
            'update_time': os.stat(summary_dir).st_mtime,
            'query_time': os.stat(summary_dir).st_ctime,
            'uncertainty_enabled': False,
        }
        self._samples = defaultdict(dict)
        self._metadata = {'explainers': [], 'metrics': [], 'labels': [], 'min_confidence': 0.5}
        self._benchmark = {'explainer_score': defaultdict(dict), 'label_score': defaultdict(dict)}

        self._status = _LoaderStatus.PENDING.value
        self._status_mutex = threading.Lock()

    @property
    def all_classes(self) -> List[Dict]:
        """
        Return a list of detailed label information, including label id, label name and sample count of each label.

        Returns:
            list[dict], a list of dict, each dict contains:

                - id (int): Label id.
                - label (str): Label name.
                - sample_count (int): Number of samples for each label.
        """
        sample_count_per_label = defaultdict(int)
        saliency_count_per_label = defaultdict(int)
        hoc_count_per_label = defaultdict(int)
        for sample in self._samples.values():
            if sample.get('image') and (sample.get('ground_truth_label') or sample.get('predicted_label')):
                for label in set(sample['ground_truth_label'] + sample['predicted_label']):
                    sample_count_per_label[label] += 1
                    if sample['inferences'][label]['saliency_maps']:
                        saliency_count_per_label[label] += 1
                    if sample['inferences'][label]['hoc_layers']:
                        hoc_count_per_label[label] += 1

        all_classes_return = [{'id': label_id,
                               'label': label_name,
                               'sample_count': sample_count_per_label[label_id],
                               'saliency_sample_count': saliency_count_per_label[label_id],
                               'hoc_sample_count': hoc_count_per_label[label_id]}
                              for label_id, label_name in enumerate(self._metadata['labels'])]
        return all_classes_return

    @property
    def query_time(self) -> float:
        """Return query timestamp of explain loader."""
        return self._loader_info['query_time']

    @query_time.setter
    def query_time(self, new_time: Union[datetime, float]):
        """
        Update the query_time timestamp manually.

        Args:
            new_time (datetime.datetime or float): Updated query_time for the explain loader.
        """
        if isinstance(new_time, datetime):
            self._loader_info['query_time'] = new_time.timestamp()
        elif isinstance(new_time, float):
            self._loader_info['query_time'] = new_time
        else:
            raise TypeError('new_time should have type of datetime.datetime or float, but receive {}'
                            .format(type(new_time)))

    @property
    def create_time(self) -> float:
        """Return the create timestamp of summary file."""
        return self._loader_info['create_time']

    @create_time.setter
    def create_time(self, new_time: Union[datetime, float]):
        """
        Update the create_time manually

        Args:
            new_time (datetime.datetime or float): Updated create_time of summary_file.
        """
        if isinstance(new_time, datetime):
            self._loader_info['create_time'] = new_time.timestamp()
        elif isinstance(new_time, float):
            self._loader_info['create_time'] = new_time
        else:
            raise TypeError('new_time should have type of datetime.datetime or float, but receive {}'
                            .format(type(new_time)))

    @property
    def explainers(self) -> List[str]:
        """Return a list of explainer names recorded in the summary file."""
        return self._metadata['explainers']

    @property
    def explainer_scores(self) -> List[Dict]:
        """
        Return evaluation results for every explainer.

        Returns:
            list[dict], A list of evaluation results of each explainer. Each item contains:

                - explainer (str): Name of evaluated explainer.
                - evaluations (list[dict]): A list of evaluation results by different metrics.
                - class_scores (list[dict]): A list of evaluation results on different labels.

                Each item in the evaluations contains:

                    - metric (str): name of metric method
                    - score (float): evaluation result

                Each item in the class_scores contains:

                    - label (str): Name of label
                    - evaluations (list[dict]): A list of evaluation results on different labels by different metrics.

                    Each item in evaluations contains:

                        - metric (str): Name of metric method
                        - score (float): Evaluation scores of explainer on specific label by the metric.
        """
        explainer_scores = []
        for explainer, explainer_score_on_metric in self._benchmark['explainer_score'].copy().items():
            metric_scores = [{'metric': metric, 'score': _round(score)}
                             for metric, score in explainer_score_on_metric.items()]
            label_scores = []
            for label, label_score_on_metric in self._benchmark['label_score'][explainer].copy().items():
                score_of_single_label = {
                    'label': self._metadata['labels'][label],
                    'evaluations': [
                        {'metric': metric, 'score': _round(score)} for metric, score in label_score_on_metric.items()
                    ],
                }
                label_scores.append(score_of_single_label)
            explainer_scores.append({
                'explainer': explainer,
                'evaluations': metric_scores,
                'class_scores': label_scores,
            })
        return explainer_scores

    @property
    def labels(self) -> List[str]:
        """Return the label recorded in the summary."""
        return self._metadata['labels']

    @property
    def metrics(self) -> List[str]:
        """Return a list of metric names recorded in the summary file."""
        return self._metadata['metrics']

    @property
    def min_confidence(self) -> Optional[float]:
        """Return minimum confidence used to filter the predicted labels."""
        return self._metadata['min_confidence']

    @property
    def sample_count(self) -> int:
        """
        Return total number of samples in the loader.

        Since the loader only return available samples (i.e. with original image data and ground_truth_label loaded in
        cache), the returned count only takes the available samples into account.

        Return:
            int, total number of available samples in the loading job.
        """
        sample_count = 0
        for sample in self._samples.values():
            if sample.get('image', False):
                sample_count += 1
        return sample_count

    @property
    def samples(self) -> List[Dict]:
        """Return the information of all samples in the job."""
        return self._samples

    @property
    def train_id(self) -> str:
        """Return ID of explain loader."""
        return self._loader_info['loader_id']

    @property
    def uncertainty_enabled(self):
        """Whether uncertainty is enabled."""
        return self._loader_info['uncertainty_enabled']

    @property
    def update_time(self) -> float:
        """Return latest modification timestamp of summary file."""
        return self._loader_info['update_time']

    @update_time.setter
    def update_time(self, new_time: Union[datetime, float]):
        """
        Update the update_time manually.

        Args:
            new_time (datetime.datetime or float): Updated time for the summary file.
        """
        if isinstance(new_time, datetime):
            self._loader_info['update_time'] = new_time.timestamp()
        elif isinstance(new_time, float):
            self._loader_info['update_time'] = new_time
        else:
            raise TypeError('new_time should have type of datetime.datetime or float, but receive {}'
                            .format(type(new_time)))

    def load(self):
        """Start loading data from the latest summary file to the loader."""
        if self.status != _LoaderStatus.LOADED.value:
            self.status = _LoaderStatus.LOADING.value

        filenames = []
        for filename in FileHandler.list_dir(self._loader_info['summary_dir']):
            if FileHandler.is_file(FileHandler.join(self._loader_info['summary_dir'], filename)):
                filenames.append(filename)
        filenames = ExplainLoader._filter_files(filenames)

        if not filenames:
            raise TrainJobNotExistError('No summary file found in %s, explain job will be delete.'
                                        % self._loader_info['summary_dir'])

        is_end = False
        while not is_end and self.status != _LoaderStatus.STOP.value:
            try:
                file_changed, is_end, event_dict = self._parser.list_events(filenames)
            except UnknownError:
                is_end = True
                break

            if file_changed:
                logger.info('Summary file in %s update, reload the data in the summary.',
                            self._loader_info['summary_dir'])
                self._clear_job()
                if self.status != _LoaderStatus.STOP.value:
                    self.status = _LoaderStatus.LOADING.value
            if event_dict:
                self._import_data_from_event(event_dict)
        self._reform_sample_info()
        if is_end:
            self.status = _LoaderStatus.LOADED.value

    @property
    def status(self):
        """Get the status of this class with lock."""
        with self._status_mutex:
            return self._status

    @status.setter
    def status(self, status):
        """Set the status of this class with lock."""
        with self._status_mutex:
            self._status = status

    def stop(self):
        """Stop load data."""
        self.status = _LoaderStatus.STOP.value

    def get_all_samples(self) -> List[Dict]:
        """
        Return a list of sample information cached in the explain job.

        Returns:
            sample_list (list[SampleObj]): a list of sample objects, each object consists of:

                - id (int): Sample id.
                - name (str): Basename of image.
                - inferences (list[dict]): List of inferences for all labels.
        """
        returned_samples = [{'id': sample_id, 'name': info['name'], 'image': info['image'],
                             'inferences': list(info['inferences'].values())} for sample_id, info in
                            self._samples.items() if info.get('image', False)]
        return returned_samples

    def _import_data_from_event(self, event_dict: Dict):
        """Parse and import data from the event data."""
        if 'metadata' not in event_dict and self._is_metadata_empty():
            raise ParamValueError('metadata is incomplete, should write metadata first in the summary.')

        for tag, event in event_dict.items():
            if tag == ExplainFieldsEnum.METADATA.value:
                self._import_metadata_from_event(event.metadata)
            elif tag == ExplainFieldsEnum.BENCHMARK.value:
                self._import_benchmark_from_event(event.benchmark)
            elif tag == ExplainFieldsEnum.SAMPLE_ID.value:
                self._import_sample_from_event(event)
            else:
                logger.info('Unknown ExplainField: %s.', tag)

    def _is_metadata_empty(self):
        """Check whether metadata is completely loaded first."""
        if not self._metadata['labels']:
            return True
        return False

    def _import_metadata_from_event(self, metadata_event):
        """Import the metadata from event into loader."""

        def take_union(existed_list, imported_data):
            """Take union of existed_list and imported_data."""
            if isinstance(imported_data, Iterable):
                for sample in imported_data:
                    if sample not in existed_list:
                        existed_list.append(sample)

        take_union(self._metadata['explainers'], metadata_event.explain_method)
        take_union(self._metadata['metrics'], metadata_event.benchmark_method)
        take_union(self._metadata['labels'], metadata_event.label)

    def _import_benchmark_from_event(self, benchmarks):
        """
        Parse the benchmark event.

        Benchmark data are separated into 'explainer_score' and 'label_score'. 'explainer_score' contains overall
        evaluation results of each explainer by different metrics, while 'label_score' additionally divides the results
        w.r.t different labels.

            The structure of self._benchmark['explainer_score'] demonstrates below:
                 {
                    explainer_1: {metric_name_1: score_1, ...},
                    explainer_2: {metric_name_1: score_1, ...},
                    ...
                 }

            The structure of self._benchmark['label_score'] is:
                {
                    explainer_1: {label_id: {metric_1: score_1, metric_2: score_2, ...}, ...},
                    explainer_2: {label_id: {metric_1: score_1, metric_2: score_2, ...}, ...},
                    ...
                }

        Args:
            benchmarks (BenchmarkContainer): Parsed benchmarks data from summary file.
        """
        explainer_score = self._benchmark['explainer_score']
        label_score = self._benchmark['label_score']

        for benchmark in benchmarks:
            explainer = benchmark.explain_method
            metric = benchmark.benchmark_method
            metric_score = benchmark.total_score
            label_score_event = benchmark.label_score

            explainer_score[explainer][metric] = _NAN_CONSTANT if math.isnan(metric_score) else metric_score
            new_label_score_dict = ExplainLoader._score_event_to_dict(label_score_event, metric)
            for label, scores_of_metric in new_label_score_dict.items():
                if label not in label_score[explainer]:
                    label_score[explainer][label] = {}
                label_score[explainer][label].update(scores_of_metric)

    def _import_sample_from_event(self, sample):
        """
        Parse the sample event.

        Detailed data of each sample are store in self._samples, identified by sample_id. Each sample data are stored
        in the following structure:

            - ground_truth_labels (list[int]): A list of ground truth labels of the sample.
            - ground_truth_probs (list[float]): A list of confidences of ground-truth label from black-box model.
            - predicted_labels (list[int]): A list of predicted labels from the black-box model.
            - predicted_probs (list[int]): A list of confidences w.r.t the predicted labels.
            - explanations (dict): Explanations is a dictionary where the each explainer name mapping to a dictionary
                of saliency maps. The structure of explanations demonstrates below:
                {
                    explainer_name_1: {label_1: saliency_id_1, label_2: saliency_id_2, ...},
                    explainer_name_2: {label_1: saliency_id_1, label_2: saliency_id_2, ...},
                    ...
                }
            - hierarchical_occlusion (dict):  A dictionary where each label is matched to a dictionary:
                {label_1: [{prob: layer1_prob, bbox: []}, {prob: layer2_prob, bbox: []}],
                 label_2:
                }
        """
        if getattr(sample, 'sample_id', None) is None:
            raise ParamValueError('sample_event has no sample_id')
        sample_id = sample.sample_id
        if sample_id not in self._samples:
            self._samples[sample_id] = {
                'id': sample_id,
                'name': str(sample_id),
                'image': sample.image_path,
                'ground_truth_label': [],
                'predicted_label': [],
                'inferences': defaultdict(dict),
                'explanation': defaultdict(dict),
                'hierarchical_occlusion': defaultdict(dict)
            }

        if sample.image_path:
            self._samples[sample_id]['image'] = sample.image_path

        for tag in _SAMPLE_FIELD_NAMES:
            if tag == ExplainFieldsEnum.GROUND_TRUTH_LABEL:
                if not self._samples[sample_id]['ground_truth_label']:
                    self._samples[sample_id]['ground_truth_label'].extend(list(sample.ground_truth_label))
            elif tag == ExplainFieldsEnum.INFERENCE:
                self._import_inference_from_event(sample, sample_id)
            elif tag == ExplainFieldsEnum.EXPLANATION:
                self._import_explanation_from_event(sample, sample_id)
            elif tag == ExplainFieldsEnum.HIERARCHICAL_OCCLUSION:
                self._import_hoc_from_event(sample, sample_id)

    def _reform_sample_info(self):
        """Reform the sample info."""
        for _, sample_info in self._samples.items():
            inferences = sample_info['inferences']
            res_dict = defaultdict(list)
            for explainer, label_heatmap_path_dict in sample_info['explanation'].items():
                for label, heatmap_path in label_heatmap_path_dict.items():
                    res_dict[label].append({'explainer': explainer, 'overlay': heatmap_path})

            for label, item in inferences.items():
                item['saliency_maps'] = res_dict[label]

            for label, item in sample_info['hierarchical_occlusion'].items():
                inferences[label]['hoc_layers'] = item['hoc_layers']

    def _import_inference_from_event(self, event, sample_id):
        """Parse the inference event."""
        inference = event.inference
        if inference.ground_truth_prob_sd or inference.predicted_prob_sd:
            self._loader_info['uncertainty_enabled'] = True
        if not self._samples[sample_id]['predicted_label']:
            self._samples[sample_id]['predicted_label'].extend(list(inference.predicted_label))
        if not self._samples[sample_id]['inferences']:
            inferences = {}
            for label, prob in zip(list(event.ground_truth_label) + list(inference.predicted_label),
                                   list(inference.ground_truth_prob) + list(inference.predicted_prob)):
                inferences[label] = {
                    'label': self._metadata['labels'][label],
                    'confidence': _round(prob),
                    'saliency_maps': [],
                    'hoc_layers': {},
                }
                if not event.ground_truth_label:
                    inferences[label]['prediction_type'] = None
                else:
                    if prob < self.min_confidence:
                        inferences[label]['prediction_type'] = 'FN'
                    elif label in event.ground_truth_label:
                        inferences[label]['prediction_type'] = 'TP'
                    else:
                        inferences[label]['prediction_type'] = 'FP'
            if self._loader_info['uncertainty_enabled']:
                for label, std, low, high in zip(
                        list(event.ground_truth_label) + list(inference.predicted_label),
                        list(inference.ground_truth_prob_sd) + list(inference.predicted_prob_sd),
                        list(inference.ground_truth_prob_itl95_low) + list(inference.predicted_prob_itl95_low),
                        list(inference.ground_truth_prob_itl95_hi) + list(inference.predicted_prob_itl95_hi)):
                    inferences[label]['confidence_sd'] = _round(std)
                    inferences[label]['confidence_itl95'] = [_round(low), _round(high)]

            self._samples[sample_id]['inferences'] = inferences

    def _import_explanation_from_event(self, event, sample_id):
        """Parse the explanation event."""
        if self._samples[sample_id]['explanation'] is None:
            self._samples[sample_id]['explanation'] = defaultdict(dict)
        sample_explanation = self._samples[sample_id]['explanation']

        for explanation_item in event.explanation:
            explainer = explanation_item.explain_method
            label = explanation_item.label
            sample_explanation[explainer][label] = explanation_item.heatmap_path

    def _import_hoc_from_event(self, event, sample_id):
        """Parse the mango event."""
        sample_hoc = self._samples[sample_id]['hierarchical_occlusion']
        if event.hierarchical_occlusion:
            for hoc_item in event.hierarchical_occlusion:
                label = hoc_item.label
                sample_hoc[label] = {}
                sample_hoc[label]['label'] = label
                sample_hoc[label]['mask'] = hoc_item.mask
                sample_hoc[label]['confidence'] = self._samples[sample_id]['inferences'][label]['confidence']
                sample_hoc[label]['hoc_layers'] = []
                for hoc_layer in hoc_item.layer:
                    sample_hoc_dict = {'confidence': hoc_layer.prob}
                    box_lst = list(hoc_layer.box)
                    box = [box_lst[i: i + 4] for i in range(0, len(hoc_layer.box), 4)]
                    sample_hoc_dict['boxes'] = box
                    sample_hoc[label]['hoc_layers'].append(sample_hoc_dict)

    def _clear_job(self):
        """Clear the cached data and update the time info of the loader."""
        self._samples.clear()
        self._loader_info['create_time'] = os.stat(self._loader_info['summary_dir']).st_ctime
        self._loader_info['update_time'] = os.stat(self._loader_info['summary_dir']).st_mtime
        self._loader_info['query_time'] = max(self._loader_info['update_time'], self._loader_info['query_time'])

        def clear_inner_dict(outer_dict):
            """Clear the inner structured data of the given dict."""
            for item in outer_dict.values():
                item.clear()

        map(clear_inner_dict, [self._metadata, self._benchmark])

    @staticmethod
    def _filter_files(filenames):
        """
        Gets a list of summary files.

        Args:
            filenames (list[str]): File name list, like [filename1, filename2].

        Returns:
            list[str], filename list.
        """
        return list(filter(
            lambda filename: (re.search(r'summary\.\d+', filename) and filename.endswith("_explain")), filenames))

    @staticmethod
    def _is_inference_valid(sample):
        """
        Check whether the inference data is empty or have the same length.

        If probs have different length with the labels, it can be confusing when assigning each prob to label.
        '_is_inference_valid' returns True only when the data size of match to each other. Note that prob data could be
        empty, so empty prob will pass the check.
        """
        ground_truth_len = len(sample['ground_truth_label'])
        for name in ['ground_truth_prob', 'ground_truth_prob_sd',
                     'ground_truth_prob_itl95_low', 'ground_truth_prob_itl95_hi']:
            if sample[name] and len(sample[name]) != ground_truth_len:
                logger.info('Length of %s not match the ground_truth_label. Length of ground_truth_label: %d,'
                            'length of %s: %d', name, ground_truth_len, name, len(sample[name]))
                return False

        predicted_len = len(sample['predicted_label'])
        for name in ['predicted_prob', 'predicted_prob_sd',
                     'predicted_prob_itl95_low', 'predicted_prob_itl95_hi']:
            if sample[name] and len(sample[name]) != predicted_len:
                logger.info('Length of %s not match the predicted_labels. Length of predicted_label: %d,'
                            'length of %s: %d', name, predicted_len, name, len(sample[name]))
                return False
        return True

    @staticmethod
    def _score_event_to_dict(label_score_event, metric) -> Dict:
        """Transfer metric scores per label to pre-defined structure."""
        new_label_score_dict = defaultdict(dict)
        for label_id, label_score in enumerate(label_score_event):
            new_label_score_dict[label_id][metric] = _NAN_CONSTANT if math.isnan(label_score) else label_score
        return new_label_score_dict



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Defaults module for mindinsight settings."""
import os

####################################
# Global default settings.
####################################
WORKSPACE = os.path.join(os.environ['HOME'], 'mindinsight')

####################################
# Web default settings.
####################################
PORT = 8080
URL_PATH_PREFIX = ''

####################################
# Debugger default settings.
####################################
DEBUGGER_PORT = 50051
ENABLE_DEBUGGER = False
# The unit is MB
OFFLINE_DEBUGGER_MEM_LIMIT = 16 * 1024
MAX_OFFLINE_DEBUGGER_SESSION_NUM = 2

####################################
# Datavisual default settings.
####################################
RELOAD_INTERVAL = 3 # Seconds
SUMMARY_BASE_DIR = os.getcwd()

# set MAX_GRAPH_NODE_SIZE to 100000, which is able to support yolov4 with one card 11 graphs
# will increase the value after supporting to load graphs in parallel
MAX_GRAPH_NODE_SIZE = 100000



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Base file system."""
from abc import ABC, abstractmethod
from collections import namedtuple
StatInfo = namedtuple("Info", ["size", "mtime"])


class BaseFileSystem(ABC):
    """Base class for file systems."""

    @abstractmethod
    def list_dir(self, path):
        """
        Abstract method for listing directories by path.

        Args:
            path (str): Directory path or file path.
        """

    @abstractmethod
    def is_dir(self, path):
        """
        Abstract method for determining if it is a directory.

        Args:
            path (str): Directory path or file path.
        """

    @abstractmethod
    def exists(self, path):
        """
        Abstract method for determining if it exists.

        Args:
            path (str): Directory path or file path.
        """

    @abstractmethod
    def file_stat(self, file_path):
        """
        Abstract method for getting file stat information.

        Args:
            file_path (str): File path.
        """

    @abstractmethod
    def join(self, path, *paths):
        """
        Abstract method for combining paths.

        Args:
            path (str): Directory path.
            *paths (str): Path or paths.
        """



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Profiler check parameters."""
def check_bool(input_param, param_name):
    """Bool type judgment."""
    if isinstance(input_param, bool):
        return input_param
    raise TypeError("Parameter {}: input type must be bool!".format(param_name))

def check_subgraph(subgraph):
    """Check subgraph."""
    if subgraph in ("all", "Default", "Gradients"):
        return subgraph
    raise ValueError("subgraph must be all or Default or Gradients, but got {}.".format(subgraph))



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The Timeline Analyser."""
import json
import os

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import ProfilerIOException
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.profiler.analyser.timeline_processor import TimelineService


class TimelineAnalyser(BaseAnalyser):
    """
    Analyse timeline data from file.
    """
    _ascend_display_filename = 'ascend_timeline_display_{}.json'
    _gpu_display_filename = 'gpu_timeline_display_{}.json'
    _ascend_timeline_summary_filename = 'ascend_timeline_summary_{}.json'
    _gpu_timeline_summary_filename = 'gpu_timeline_summary_{}.json'

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def get_display_timeline(self, device_type, scope_name_num):
        """
        Get timeline data for UI display.

        Returns:
            json, the content of timeline data.
        """
        if device_type == "ascend":
            display_filename = self._ascend_display_filename.format(self._device_id)
        elif device_type == "gpu":
            display_filename = self._gpu_display_filename.format(self._device_id)
        else:
            logger.info('device type should be ascend or gpu. Please check the device type.')
            raise ParamValueError("Invalid device_type.")
        file_path = os.path.join(self._profiling_dir, display_filename)
        file_path = validate_and_normalize_path(
            file_path, raise_key='Invalid timeline json path.'
        )

        timeline = []
        if os.path.exists(file_path):
            try:
                with open(file_path, 'r') as f_obj:
                    timeline = json.load(f_obj)
                    for idx, time_item in enumerate(timeline):
                        if time_item.get("tid") == 100001 and \
                                time_item["ph"] != "M" and \
                                int(time_item["scope_level"]) >= int(scope_name_num):
                            timeline[idx] = None
                    timeline = list(filter(lambda x: x, timeline))
            except (IOError, OSError, json.JSONDecodeError) as err:
                logger.error('Error occurred when read timeline display file: %s', err)
                raise ProfilerIOException()
        else:
            logger.info('No timeline file. Please check the output path.')

        return timeline

    def get_timeline_summary(self, device_type):
        """
        Get timeline summary information for UI display.

        Returns:
            json, the content of timeline summary information.
        """
        if device_type == "ascend":
            summary_filename = self._ascend_timeline_summary_filename.format(self._device_id)
        elif device_type == "gpu":
            summary_filename = self._gpu_timeline_summary_filename.format(self._device_id)
        else:
            logger.info('device type should be ascend or gpu. Please check the device type.')
            raise ParamValueError("Invalid device_type.")
        file_path = os.path.join(self._profiling_dir, summary_filename)
        file_path = validate_and_normalize_path(
            file_path, raise_key='Invalid timeline summary path.'
        )

        timeline_summary = {}
        if os.path.exists(file_path):
            try:
                with open(file_path, 'r') as f_obj:
                    timeline_summary = json.load(f_obj)
            except (IOError, OSError, json.JSONDecodeError) as err:
                logger.error('Error occurred when read timeline summary file: %s', err)
                raise ProfilerIOException()
        else:
            logger.info('No timeline summary file. Please check the output path.')

        timeline_summary.setdefault("max_scope_name_num", 0)

        return timeline_summary

    def get_marey_timeline(self, step, device_list):
        operator_time_maps, min_time, max_time, stage_data = TimelineService(self._profiling_dir,
                                                                             device_list).get_ops_by_step(step)
        ret = {"maps": operator_time_maps, "minT": min_time, "maxT": max_time, "stage_data": stage_data}
        return ret



# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Base module."""

import os
import re
import enum
import collections

import numpy as np

from mindinsight.domain.graph.exceptions import UnknownTensorError


class MindSporeType(enum.Enum):
    """MindSpore Type."""
    INT = 'int32'
    UINT = 'uint'
    FLOAT = 'float16'
    TENSOR = 'tensor'


class DeviceType(enum.Enum):
    """Device Type."""
    ASCEND = 'ascend'
    GPU = 'gpu'


class DumpType(enum.Enum):
    """Dump Type."""
    E2E = 'e2e'
    ASYNC = 'async'


class Tensor:
    """
    Tensor object of dump file.

    Args:
        op_id (str): Operator ID.
        index (int): Index of operator inputs/outputs.
        file_path (str): Absolute file path of tensor file.
    """

    FEATURE = collections.namedtuple('Feature', ['type', 'id', 'io'])
    VALUE = collections.namedtuple('Value', ['index', 'shape', 'dtype', 'path'])

    @classmethod
    def extract_shape_from_str(cls, shape_str):
        """
        Extract shape from tensor file.

        Args:
            shape_str (str): Shape string.

        Returns:
            tuple, shape of tensor file.
        """
        shape = tuple([int(dim.strip()) for dim in shape_str.strip('_').split('_')])
        # The shape info in dump file name is (0,) which is inconsistent with the actual tensor shape.
        # The shape needs to be converted to (1,).
        if shape == (0,):
            shape = (1,)
        return shape

    @classmethod
    def parse_tensor_file_name(cls, file_name):
        """
        Parse tensor file name.

        Args:
            file_name (str): Tensor file name.

        Returns:
            bool, indicating if node is operator.
            dict, tensor file info.

        Raises:
            UnknownTensorError: If tensor file name can not be recognized.
        """
        is_op = False
        is_npy = file_name.endswith('.npy')
        if re.search(r'-op\d+(_|(\.\d+\.\d+\.))(input|output)(_|\.)', file_name):
            is_op = True
            dump_type = DumpType.E2E
            if re.search(r'-op(?P<op_id>\d+)\.(?P<stream_id>\d+)\.(?P<task_id>\d+)', file_name):
                dump_type = DumpType.ASYNC

            if dump_type == DumpType.ASYNC:
                file_name = file_name[file_name.find('.')+1:]
                if is_npy:
                    regex = r'_(?P<op_name>[A-Za-z0-9]+)-op(?P<op_id>\d+)' \
                            r'\.(?P<stream_id>\d+)\.(?P<task_id>\d+)' \
                            r'\.(?P<io>input|output)' \
                            r'\.(?P<index>\d+)' \
                            r'\.npy$'
                else:
                    regex = r'_(?P<op_name>[A-Za-z0-9]+)-op(?P<op_id>\d+)' \
                            r'\.(?P<stream_id>\d+)\.(?P<task_id>\d+)' \
                            r'\.(?P<io>input|output)' \
                            r'\.(?P<index>\d+)' \
                            r'\.(?P<shape>[0-9\_]+)' \
                            r'\.(?P<dtype>bool|((uint|int|float)\d+))' \
                            r'\.(?P<format>[A-Za-z0-9\_]+)\.bin$'

            else:
                regex = r'--(?P<op_name>[A-Za-z0-9\_]+)-op(?P<op_id>\d+)' \
                        r'_(?P<io>input|output)' \
                        r'_(?P<index>\d+)' \
                        r'_shape_(?P<shape>[0-9\_]+)' \
                        r'_.*(?P<dtype>Bool|((UInt|Int|Float)\d+))' \
                        r'_(?P<format>[A-Za-z0-9\_]+)\.bin$'

        else:
            regex = r'^(?P<node_name>[A-Za-z0-9\.\_]+)' \
                    r'_(?P<io>input|output)' \
                    r'_(?P<index>\d+)' \
                    r'_shape_(?P<shape>[0-9\_]+)' \
                    r'_.*(?P<dtype>Bool|((UInt|Int|Float)\d+))' \
                    r'_(?P<format>[A-Za-z0-9\_]+)\.bin$'

        pattern = re.search(regex, file_name)
        if pattern is None:
            raise UnknownTensorError(is_op, file_name)

        info = pattern.groupdict()
        info['index'] = int(info['index'])
        info['shape'] = None if is_npy else cls.extract_shape_from_str(info['shape'])
        info['dtype'] = None if is_npy else info['dtype'].lower()
        return is_op, info

    @classmethod
    def scan_tensors(cls, tensor_dir):
        """
        Scan tensors.

        Args:
            tensor_dir (str): Directory path where holds the tensor files.
            check (lambda): Function to check tensor values.

        Returns:
            dict, tensor file mapping.
        """
        tensor_mapping = {}
        if not tensor_dir:
            return tensor_mapping

        file_names = os.listdir(tensor_dir)
        for file_name in file_names:
            full_path = os.path.join(tensor_dir, file_name)
            if not re.search(r'\.(bin|npy)$', file_name) or os.path.isdir(full_path):
                continue

            try:
                is_op, info = cls.parse_tensor_file_name(file_name)
            except UnknownTensorError:
                continue

            if is_op:
                feature = cls.FEATURE(type=info['op_name'], id=info['op_id'], io=info['io'])
            else:
                feature = cls.FEATURE(type='', id=info['node_name'], io=info['io'])

            value = cls.VALUE(index=info['index'], shape=info['shape'], dtype=info['dtype'], path=full_path)
            tensors = tensor_mapping.get(feature)
            if tensors:
                tensor_mapping[feature].append(value)
                tensor_mapping[feature].sort(key=lambda x: x[0])
            else:
                tensor_mapping[feature] = [value]

        return tensor_mapping

    def __init__(self, op_id, index, file_path):
        self.op_id = op_id
        self.index = index
        self.file_path = file_path

    def load(self):
        """
        Load tensor file.

        Returns:
            ndarray, tensor data.
        """
        if self.file_path.endswith('.npy'):
            tensor = np.load(self.file_path)
            return tensor

        metas = self.metas
        if metas is None:
            return None
        dtype = getattr(np, metas['dtype'])
        tensor = np.fromfile(self.file_path, dtype=dtype)
        try:
            tensor = tensor.reshape(metas['shape'])
        except ValueError:
            pass
        return tensor

    @property
    def metas(self):
        """
        Metas property.

        Returns:
            dict, metas extracted from tensor file name.
        """
        file_name = os.path.basename(self.file_path)
        try:
            is_op, info = self.parse_tensor_file_name(file_name)
        except UnknownTensorError:
            return None

        if is_op:
            info.pop('op_name')
            info.pop('op_id')
        else:
            info.pop('node_name')

        if file_name.endswith('.npy'):
            info.pop('dtype')
            info.pop('shape')

        return info

    @property
    def full_name(self):
        """
        Full name property.

        Returns:
            str, full name.
        """
        full_name_str, _ = os.path.basename(self.file_path).split('_output_')
        return full_name_str.replace('--', '/')

    @property
    def scope(self):
        """
        Scope property.

        Returns:
            str, scope.
        """
        return os.path.dirname(self.full_name)

    def __repr__(self):
        return str({
            'op_id': self.op_id,
            'index': self.index,
            'file_path': self.file_path,
        })


class NodeType(enum.Enum):
    """Node Type."""
    OPERATOR = 'operator'
    PARAMETER = 'parameter'
    CONSTANT = 'constant'


class InputType(enum.Enum):
    """Input Type."""
    OPERATOR = 'operator'
    PARAMETER = 'parameter'
    CONSTANT = 'constant'
    TENSOR = 'tensor'
    SCALAR = 'scalar'
    REFERENCE = 'reference'
    NONE = 'none'


class OutputType(enum.Enum):
    """Output Type."""
    NONE = 'none'
    BOOL = 'bool'
    INT8 = 'int8'
    INT16 = 'int16'
    INT32 = 'int32'
    INT64 = 'int64'
    UINT8 = 'uint8'
    UINT16 = 'uint16'
    UINT32 = 'uint32'
    UINT64 = 'uint64'
    FLOAT16 = 'float16'
    FLOAT32 = 'float32'
    FLOAT64 = 'float64'
    TENSOR = 'tensor'
    TUPLE = 'tuple'
    STRING = 'string'


class NodeInput:
    """
    Graph node input.

    Args:
        input_type (InputType): Input type.
        input_name (str): Input name.
    """

    def __init__(self, input_type, input_name):
        self.type = input_type
        self.name = input_name
        self.op_id = ''
        self.info = None

    def __repr__(self):
        return str({
            'type': self.type,
            'name': self.name,
            'op_id': self.op_id,
            'info': self.info,
        })


class NodeOutput:
    """
    Graph node output.

    Args:
        output_type (OutputType): Output type.
    """

    SCALAR_TYPES = (
        OutputType.INT8,
        OutputType.INT16,
        OutputType.INT32,
        OutputType.INT64,
        OutputType.UINT8,
        OutputType.UINT16,
        OutputType.UINT32,
        OutputType.UINT64,
        OutputType.FLOAT16,
        OutputType.FLOAT32,
    )

    def __init__(self, output_type):
        self.type = output_type
        if output_type == OutputType.BOOL:
            self.info = dict(value=None)
            self.slot_size = 1
        elif output_type == OutputType.STRING:
            self.info = dict(value=None)
            self.slot_size = 1
        elif output_type in self.SCALAR_TYPES:
            self.info = dict(value=None)
            self.slot_size = 1
        elif output_type == OutputType.TENSOR:
            self.info = dict(dtype='', shape=(), tensor=None)
            self.slot_size = 1
        elif output_type == OutputType.TUPLE:
            self.info = dict(dtypes=[], shapes=[], tensors=[])
            # need update when parse output items
            self.slot_size = 0
        else:
            self.info = None
            self.slot_size = 1

    def __repr__(self):
        return str({
            'type': self.type,
            'info': self.info,
        })


class Source:
    """
    Source address info.

    Args:
        file_path (str): Absolute path of source file.
        line_no (int): Line number of code line in source file.
        code_line (int): Code line content.
    """

    def __init__(self, file_path, line_no, code_line, has_substack):
        self.file_path = file_path
        self.line_no = line_no
        self.code_line = code_line
        self.has_substack = has_substack

    def to_dict(self):
        """Parse to dict."""
        return {
            'file_path': self.file_path,
            'line_no': self.line_no,
            'code_line': self.code_line,
            'has_substack': self.has_substack,
        }

    def __repr__(self):
        return str(self.to_dict())

    def __str__(self):
        if not self.file_path:
            return self.code_line
        substack_symbol = '-' if self.has_substack else ' '
        return f'# {substack_symbol} {self.file_path}:{self.line_no}  {self.code_line}'

    @classmethod
    def build_stack_from_source_address(cls, source_address):
        """
        Build stack from source address.

        Args:
            source_address (str): Source address content.

        Returns:
            list, list of Source objects.
        """
        stack = []
        for line in source_address.strip().split('\n'):
            regex = r'#(\s\s(?P<has_substack>\s|-))?\sIn\sfile\s(?P<file_path>.+)\:(?P<line_no>\d+)/(?P<code_line>.+)/'
            pattern = re.search(regex, line.strip())
            if pattern is None:
                source = {
                    'file_path': '',
                    'line_no': 0,
                    'code_line': line.strip(),
                    'has_substack': False,
                }
            else:
                source = pattern.groupdict()
                source['line_no'] = int(source['line_no'])
                source['code_line'] = source['code_line'].strip()
                source['has_substack'] = source['has_substack'] == '-'
            stack.append(cls(**source))

        return stack


class Node:
    """
    Graph node.

    Args:
        raw (NodeProto): Node Proto.
    """

    def __init__(self, raw):
        self.raw = raw
        self.name = ''
        self.output = None
        self.downstream = []


class Constant(Node):
    """Constant node within graph."""

    def __init__(self, raw):
        super().__init__(raw)
        self.name = raw.key
        self.full_name = raw.full_name

    def __repr__(self):
        return str({
            'name': self.name,
            'output': self.output,
            'downstream': self.downstream,
            'full_name': self.full_name
        })


class Parameter(Node):
    """Parameter node within graph."""

    def __init__(self, raw):
        super().__init__(raw)
        self.name = raw.name

    def __repr__(self):
        return str({
            'name': self.name,
            'output': self.output,
            'downstream': self.downstream,
        })


class Operator(Node):
    """Operator node within graph."""

    def __init__(self, raw):
        super().__init__(raw)
        self.name = raw.name
        self.type = raw.op_type
        self.scope = raw.scope or os.path.dirname(raw.full_name)
        self.full_name = raw.full_name
        self.inputs = []
        self.attrs = {}
        self.stack = []

    @property
    def op_id(self):
        """
        Op ID property.

        Returns:
            str, op ID.
        """
        pattern = re.search(r'-op(?P<op_id>\d+)$', self.full_name)
        if not pattern:
            return self.name

        info = pattern.groupdict()
        return info['op_id']

    def __repr__(self):
        return str({
            'name': self.name,
            'type': self.type,
            'inputs': self.inputs,
            'output': self.output,
            'downstream': self.downstream,
            'attrs': self.attrs,
            'scope': self.scope,
            'full_name': self.full_name,
            'op_id': self.op_id,
        })


class Parser:
    """Graph file parser."""

    def __init__(self, graph_data=None, tensor_dir=''):
        self.graph_data = graph_data
        self.tensor_dir = os.path.realpath(tensor_dir) if tensor_dir else ''

        self.constants = []
        self.parameters = []
        self.operators = []
        self.tensor_mapping = {}

    def parse(self):
        """Parse."""
        raise NotImplementedError


class NodeTypeEnum(enum.Enum):
    """Node type enum. The following types are new to our custom."""
    NAME_SCOPE = 'name_scope'
    AGGREGATION_SCOPE = 'aggregation_scope'
    PARAMETER = 'Parameter'
    CONST = 'Const'
    LOAD = 'Load'
    MAKETUPLE = 'MakeTuple'
    TUPLE_GET_ITEM = 'TupleGetItem'
    UPDATE_STATE = 'UpdateState'


class DebuggerSource(Source):
    """Source Data object"""

    @property
    def stack(self):
        """The property of stack."""
        return [self]

    def __lt__(self, other):
        pkg_pattern = 'site-packages'
        cur_path_value = int(pkg_pattern in self.file_path)
        other_path_value = int(pkg_pattern in other.file_path)
        if cur_path_value != other_path_value:
            return cur_path_value < other_path_value
        if self.file_path != other.file_path:
            return self.file_path < other.file_path
        return self.line_no < other.line_no

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    def __hash__(self):
        return hash(str(self.to_dict()))



# Copyright 2021-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The cpu base analyser."""
import csv
import os

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.validator import validate
from mindinsight.profiler.common.exceptions.exceptions import ProfilerRawFileException
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path


class CpuAnalyser(BaseAnalyser):
    """Cpu base analyser."""
    _csv_file_to_analyse = ""

    def _load(self):
        """Load data according to the parsed CPU operator types file."""
        op_type_file_path = os.path.join(
            self._profiling_dir,
            self._csv_file_to_analyse.format(self._device_id)
        )
        op_type_file_path = validate_and_normalize_path(
            op_type_file_path, raise_key="Invalid op_type_file_path")
        if not os.path.isfile(op_type_file_path):
            log.warning('The file <%s> does not exist.', op_type_file_path)
            return

        with open(op_type_file_path, 'r') as file:
            csv_reader = csv.reader(file)
            _ = next(csv_reader)
            for info in csv_reader:
                self._data.append(self._convert_field_type(info))

    @staticmethod
    def _convert_field_type(row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        return row

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """
        def _inner_filter(item: list):
            return self._default_filter(item, filter_condition)

        self._result = list(filter(_inner_filter, self._data))

class CpuOpTypeAnalyser(CpuAnalyser):
    """Cpu operation type analyser."""
    _col_names = validate.CPU_TYPE_COL
    _csv_file_to_analyse = 'cpu_op_type_info_{}.csv'

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        try:
            return [row[0], int(row[1]), int(row[2]),
                    self._format_float_data(float(row[3]) * self._ms_to_us),
                    self._format_float_data(float(row[4]) * self._ms_to_us),
                    self._format_float_data(float(row[5])*100)]
        except IndexError as err:
            log.exception(err)
            raise ProfilerRawFileException('failed to get HOST CPU operator type data.')


class CpuOpInfoAnalyser(CpuAnalyser):
    """Cpu operation detail info analyser."""
    _col_names = validate.CPU_DETAIL_COL
    _csv_file_to_analyse = 'cpu_op_detail_info_{}.csv'

    def _convert_field_type(self, row):
        """
        Convert the field type to the specific type.

        Args:
            row (list): One row data from parsed data.

        Returns:
            list, the converted data.
        """
        try:
            return [row[0], row[1], row[2], row[3], int(row[4]),
                    self._format_float_data(float(row[5]) * self._ms_to_us),
                    self._format_float_data(float(row[6]) * self._ms_to_us),
                    self._format_float_data(float(row[7])), row[8]]
        except IndexError as err:
            log.exception(err)
            raise ProfilerRawFileException('failed to get HOST CPU operator detail data.')



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Trigger data manager load."""
import time

from mindinsight.conf import settings
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.data_transform.data_manager import DATA_MANAGER
from mindinsight.lineagemgr.cache_item_updater import LineageCacheItemUpdater
from mindinsight.debugger.debugger_folder_analyzer import DebuggerFolderAnalyzer

ANALYZERS = {
    "debugger_folder_analyzer": DebuggerFolderAnalyzer()
}


def init_module(app):
    """
    Interface to init module.

    Args:
        app (Flask): An instance of Flask.

    """
    # Just to suppress pylint warning about unused arg.
    logger.debug("App: %s", type(app))
    for analyzer in ANALYZERS.values():
        DATA_MANAGER.register_folder_analyzer(analyzer)
    DATA_MANAGER.register_brief_cache_item_updater(LineageCacheItemUpdater())
    # Let gunicorn load other modules first.
    time.sleep(1)

    DATA_MANAGER.start_load_data(reload_interval=settings.RELOAD_INTERVAL)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""General tuner."""
import json
import os
import shlex
import subprocess
import uuid
import yaml

from marshmallow import ValidationError

from mindinsight.lineagemgr.common.validator.validate_path import safe_normalize_path
from mindinsight.lineagemgr.model import get_lineage_table, LineageTable
from mindinsight.optimizer.common.constants import HYPER_CONFIG_ENV_NAME
from mindinsight.optimizer.common.enums import TuneMethod
from mindinsight.optimizer.common.exceptions import OptimizerTerminateError, ConfigParamError
from mindinsight.optimizer.common.log import logger
from mindinsight.optimizer.common.validator.optimizer_config import OptimizerConfig
from mindinsight.optimizer.tuners.gp_tuner import GPBaseTuner
from mindinsight.optimizer.utils.param_handler import organize_params_target
from mindinsight.optimizer.utils.utils import get_nested_message
from mindinsight.utils.exceptions import MindInsightException, ParamValueError, FileSystemPermissionError, UnknownError

_OK = 0


class Tuner:
    """
    Tuner for auto tuning.

    Args:
        config_path (str): config path, a yaml format file containing settings about tuner, target and parameters, etc.

    Raises:
        FileSystemPermissionError, can not open the config file because of permission.
        UnknownError, other exception.
    """
    def __init__(self, config_path: str):
        self._config_info = self._validate_config(config_path)
        self._summary_base_dir = self._config_info.get('summary_base_dir')
        self._dir_prefix = 'train'

    def _validate_config(self, config_path):
        """Check config_path."""
        config_path = self._normalize_path("config_path", config_path)
        try:
            with open(config_path, "r") as file:
                config_info = yaml.safe_load(file)
        except PermissionError as exc:
            raise FileSystemPermissionError("Can not open config file. Detail: %s." % str(exc))
        except Exception as exc:
            raise UnknownError("Detail: %s." % str(exc))

        self._validate_config_schema(config_info)
        config_info['summary_base_dir'] = self._normalize_path("summary_base_dir", config_info.get('summary_base_dir'))
        self._make_summary_base_dir(config_info['summary_base_dir'])
        return config_info

    def _validate_config_schema(self, config_info):
        error = OptimizerConfig().validate(config_info)
        if error:
            err_msg = get_nested_message(error)
            raise ConfigParamError(err_msg)

    def _make_summary_base_dir(self, summary_base_dir):
        """Check and make summary_base_dir."""
        if not os.path.exists(summary_base_dir):
            permissions = os.R_OK | os.W_OK | os.X_OK
            os.umask(permissions << 3 | permissions)
            mode = permissions << 6
            try:
                logger.info("The summary_base_dir is generated automatically, path is %s.", summary_base_dir)
                os.makedirs(summary_base_dir, mode=mode, exist_ok=True)
            except OSError as exc:
                raise UnknownError("Can not make the summary base directory. Detail: %s." % str(exc))

    def _normalize_path(self, param_name, path):
        """Normalize config path."""
        path = os.path.realpath(path)
        try:
            path = safe_normalize_path(
                path, param_name, None, check_absolute_path=True
            )
        except ValidationError:
            logger.error("The %r is invalid.", param_name)
            raise ParamValueError("The %r is invalid." % param_name)

        return path

    def _update_from_lineage(self):
        """Update lineage from lineagemgr."""
        try:
            lineage_table = get_lineage_table(summary_base_dir=self._summary_base_dir)
        except MindInsightException as err:
            logger.info("Can not query lineage. Detail: %s", str(err))
            lineage_table = None

        self._lineage_table = lineage_table

    def optimize(self, max_expr_times=1):
        """Method for auto tuning."""
        target_info = self._config_info.get('target')
        params_info = self._config_info.get('parameters')
        command = self._config_info.get('command')
        tuner = self._config_info.get('tuner')
        for _ in range(max_expr_times):
            self._update_from_lineage()
            suggestion, user_defined_info = self._suggest(self._lineage_table, params_info, target_info, tuner)

            hyper_config = {
                'params': suggestion,
                'summary_dir': os.path.join(self._summary_base_dir, f'{self._dir_prefix}_{str(uuid.uuid1())}'),
                'custom_lineage_data': user_defined_info
            }
            logger.info("Suggest values are: %s.", suggestion)
            os.environ[HYPER_CONFIG_ENV_NAME] = json.dumps(hyper_config)
            s = subprocess.Popen(shlex.split(command))
            s.wait()
            if s.returncode != _OK:
                logger.error("An error occurred during execution, the auto tuning will be terminated.")
                raise OptimizerTerminateError("An error occurred during execution, the auto tuning was terminated.")

    def _get_tuner(self, tuner):
        """Get tuner."""
        if tuner is None:
            return GPBaseTuner()

        tuner_name = tuner.get("name").lower()
        if tuner_name not in TuneMethod.list_members():
            raise ParamValueError("'tune_method' should in %s." % TuneMethod.list_members())

        args = tuner.get("args")
        if args is not None and args.get("method") is not None:
            return GPBaseTuner(args.get("method"))

        # Only support gaussian process regressor currently.
        return GPBaseTuner()

    def _suggest(self, lineage_table: LineageTable, params_info: dict, target_info: dict, tuner):
        """Get suggestions for targets."""
        tuner = self._get_tuner(tuner)
        param_matrix, target_column = organize_params_target(lineage_table, params_info, target_info)
        suggestion, user_defined_info = tuner.suggest(param_matrix, target_column, params_info)

        return suggestion, user_defined_info



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The specific cluster analyser class"""
import json
import os
import csv
import re
import pandas as pd

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import ProfilerFileNotFoundException, \
    ProfilerDirNotFoundException, ProfilerIOException
from mindinsight.profiler.common.log import logger as log
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.util import analyse_device_list_from_profiler_dir


class ClusterAnalyser(BaseAnalyser):
    """The analyser for analyzing the cluster."""

    def __init__(self, cluster_profiler_dir, device_id):
        super().__init__(cluster_profiler_dir, device_id)
        self._cluster_profiler_dir = cluster_profiler_dir
        self._target_dir_path = self._get_target_dir_path()
        self._cluster_rank_ids = self._get_cluster_rank_ids()

    def _get_target_dir_path(self):
        """Get the target directory path."""
        target_dir_path = self._cluster_profiler_dir
        target_dir_path = validate_and_normalize_path(
            target_dir_path, raise_key="Invalid profiler dir path.")
        if not os.path.exists(target_dir_path):
            log.error('Did not find cluster_profiler dir : %s', target_dir_path)
            raise ProfilerDirNotFoundException(msg='Did not find cluster_profiler dir:{}'.format(target_dir_path))

        return target_dir_path

    def _get_cluster_rank_ids(self):
        """Get the logical card number list in cluster training."""
        cluster_rank_ids = []
        entries = os.scandir(self._target_dir_path)
        for entry in entries:
            if entry.is_symlink():
                continue
            if entry.is_file() and entry.name.startswith('step_trace_raw'):
                # index3：rank_id
                rank_id = entry.name.split("_")[3]
                if not rank_id.isdigit():
                    log.error("The value of rank id should be a number.")
                    raise TypeError("The value of rank id should be a number.")
                cluster_rank_ids.append(int(rank_id))
        cluster_rank_ids.sort()
        if cluster_rank_ids:
            # Judge whether the rank ids are continuous.
            if (cluster_rank_ids[-1] - cluster_rank_ids[0]) != len(cluster_rank_ids) - 1:
                log.warning("The rank ids are not continuous，"
                            "please check for missing files. Rank ids: %s", cluster_rank_ids)
        return cluster_rank_ids

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """


class ClusterStepTraceAnalyser(ClusterAnalyser):
    """The analyser for analyzing the cluster step trace."""

    def __init__(self, cluster_profiler_dir, device_id):
        super().__init__(cluster_profiler_dir, device_id)
        self._none_sort_col_names = []
        self._device_type, self._parallel_mode, self._stage_num, self._rank_size = \
            self._get_parallel_context()
        self._col_names = self._get_col_names()
        self._total_step_num = self._get_total_step_num()

    def _get_parallel_context(self):
        """Get the parallel mode from file name."""
        parallel_mode = "data-parallel"
        stage_num = 1
        rank_size = 1
        device_type = None
        for filename in os.listdir(self._target_dir_path):
            if filename.startswith('ascend_cluster_analyse') or filename.startswith('gpu_cluster_analyse'):
                # the format of filename is
                # 'ascend_cluster_analyse_{parallel-mode}_{stage_num}_{rank_size}_{rank_id}.csv'
                device_type = filename.split('_')[0]
                parallel_mode = filename.split('_')[3]
                stage_num = int(filename.split('_')[4])
                rank_size = int(filename.split('_')[5])

        return device_type, parallel_mode, stage_num, rank_size

    def _get_col_names(self):
        """Get the column names depends on parallel mode."""
        col_names = ['iteration_interval', 'fp_and_bp', 'tail']
        if self._parallel_mode == 'model-parallel':
            col_names = ['iteration_interval', 'computation', 'communication_alone']
        elif self._parallel_mode == 'pipeline-parallel':
            col_names = ['iteration_interval', 'computation', 'communication_alone', 'stage', 'receive_alone',
                         'collective_communication_alone']

        return col_names

    def _get_total_step_num(self):
        """Get the num of train step."""
        total_step_num = 0

        entries = os.scandir(self._target_dir_path)
        for entry in entries:
            if entry.is_symlink():
                continue
            if entry.is_file() and entry.name.startswith('step_trace_raw'):
                file_path = os.path.join(self._target_dir_path, entry.name)
                with open(file_path, 'r') as src_file:
                    lines = src_file.readlines()
                # The penultimate line represents the information of the last step
                # The step num index is 0
                if len(lines) > 1:
                    total_step_num = lines[-2].split(',')[0]
                break
        return total_step_num

    def _get_cluster_step_trace_info(self, step_num):
        """Get cluster step trace info."""
        cluster_step_trace_info = list()
        for item in self._cluster_rank_ids:
            step_trace_info = self._get_step_trace_info(item, step_num)
            step_trace_info.append(item)
            cluster_step_trace_info.append(step_trace_info)
        self._cluster_info_size = len(cluster_step_trace_info)
        return cluster_step_trace_info

    def _get_step_trace_info(self, rank_id, step_num):
        """Get step trace info."""
        file_name = 'step_trace_raw_{}_detail_time.csv'.format(rank_id)
        step_trace_file_path = \
            os.path.join(self._target_dir_path, file_name)
        step_trace_file_path = validate_and_normalize_path(
            step_trace_file_path, raise_key="Invalid step trace file path.")
        if not os.path.exists(step_trace_file_path):
            log.error('Did not find the file: %s', step_trace_file_path)
            raise ProfilerFileNotFoundException(msg='Did not find the file:{}'.format(step_trace_file_path))

        with open(step_trace_file_path, 'r') as src_file:
            lines = src_file.readlines()
            # when the step_num value is 0, it means the average value.
            # The last line of the step_trace_raw_{}_detail_time.csv records the average value.
            if step_num == 0:
                step_trace_info = lines[-1].strip('\n').split(',')
            else:
                step_trace_info = lines[step_num].strip('\n').split(',')
        # step_trace_info[6]: iteration_interval time
        # step_trace_info[7]: fp_and_bp time
        # step_trace_info[8]: tail time
        # divided by 1e5, the unit becomes a millisecond
        iter_total_time = round(float(step_trace_info[3]) / 1e5, 4)
        iteration_interval = round(float(step_trace_info[6]) / 1e5, 4)
        fp_and_bp = round(float(step_trace_info[7]) / 1e5, 4)
        tail = round(float(step_trace_info[8]) / 1e5, 4)
        step_trace_info = [iteration_interval, fp_and_bp, tail, iter_total_time]
        return step_trace_info

    def _get_cluster_step_bottleneck_info(self, step_num, stage_id):
        """Get cluster step bottleneck info."""
        cluster_step_bottleneck_info = []
        for rank_id in self._cluster_rank_ids:
            cur_stage_id = int(int(rank_id) / int((self._rank_size / self._stage_num))) + 1
            # If stage_id is 0 (default value), display all data.
            if stage_id not in (0, cur_stage_id):
                continue
            step_bottleneck_info = self.get_step_bottleneck_info(rank_id, step_num)
            step_bottleneck_info.append(rank_id)
            cluster_step_bottleneck_info.append(step_bottleneck_info)
        self._cluster_info_size = len(cluster_step_bottleneck_info)
        return cluster_step_bottleneck_info

    def get_step_bottleneck_info(self, rank_id, step_num):
        """Get cluster analyse info."""
        if self._device_type == 'ascend':
            file_name = \
                f'ascend_cluster_analyse_{self._parallel_mode}_{self._stage_num}_{self._rank_size}_{rank_id}.csv'
        else:
            file_name = \
                f'gpu_cluster_analyse_{self._parallel_mode}_{self._stage_num}_{self._rank_size}_{rank_id}.csv'
        step_bottleneck_file_path = os.path.join(self._target_dir_path, file_name)
        step_bottleneck_file_path = validate_and_normalize_path(
            step_bottleneck_file_path, raise_key="Invalid step trace file path.")
        if not os.path.exists(step_bottleneck_file_path):
            log.error('Did not find the file: %s', step_bottleneck_file_path)
            raise ProfilerFileNotFoundException(msg='Did not find the file:{}'.format(step_bottleneck_file_path))

        with open(step_bottleneck_file_path, 'r') as src_file:
            lines = src_file.readlines()
            # when the step_num value is 0, it means the average value.
            # The last line of the ascend_cluster_analyse_xxx.csv records the average value.
            if step_num == 0:
                step_bottleneck_info = lines[-1].strip('\n').split(',')
            else:
                step_bottleneck_info = lines[step_num].strip('\n').split(',')

        step_trace_info = self._get_step_trace_info(rank_id, step_num)
        # Insert the step iteration_interval time into the cluster_bottleneck_info
        step_bottleneck_info.insert(0, str(step_trace_info[0]))
        step_bottleneck_info = list(map(float, step_bottleneck_info))

        return step_bottleneck_info

    def get_overview_time_info(self):
        """Get overview time info."""
        device_list, _, _ = analyse_device_list_from_profiler_dir(self._profiling_dir)
        data = {}
        for device in device_list:
            file_name = 'step_trace_raw_{}_detail_time.csv'.format(device)
            file_path = os.path.join(self._profiling_dir, file_name)
            file_path = validate_and_normalize_path(
                file_path, raise_key='Invalid step trace file path.'
            )
            if not os.path.exists(file_path):
                log.error('Did not find the file: %s', file_path)
                raise ProfilerFileNotFoundException(msg='Did not find the file:{}'.format(file_path))
            df = pd.read_csv(file_path)
            df = pd.DataFrame(df, columns=['step_num', 'start_point', 'end_point', 'total', \
                                           'fp_point', 'bp_point', 'iteration_interval', 'fp_and_bp', 'tail'])
            cur_data = list(json.loads(df.to_json(orient="index")).values())
            device_entry = "device" + device
            data[device_entry] = cur_data
        return data

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.

                - step_id (int): The selected  step id.
        """

        step_id = int(filter_condition.get("step_id", 0))
        stage_id = int(filter_condition.get("stage_id", 0))
        if self._parallel_mode == "data-parallel":
            self._result = self._get_cluster_step_trace_info(step_id)
        else:
            self._result = self._get_cluster_step_bottleneck_info(step_id, stage_id)

    def _organize_query_result(self):
        """
        Organize the query result.

        Returns:
            dict, the query result.
        """
        result = list()
        # item[:-1] info depends on parallel mode, item[-1]:rank_id
        for item in self._result:
            info = dict()
            if self._parallel_mode == "data-parallel":
                info["step_trace_info"] = item[:-1]
            else:
                info["step_bottleneck_info"] = item[:-1]
            info["rank_id"] = item[-1]
            info["profiler_dir"] = 'profiler'
            result.append(info)

        return {
            'total_step_num': self._total_step_num,
            'stage_num': self._stage_num,
            'info': result,
            'size': self._cluster_info_size,
            'parallel-mode': self._parallel_mode
        }


class ClusterMemoryAnalyser(ClusterAnalyser):
    """The analyser for analyzing the cluster memory usage."""
    _summary_filename = 'memory_usage_summary_{}.json'

    def __init__(self, cluster_profiler_dir, device_id='0'):
        super().__init__(cluster_profiler_dir, device_id)

    def get_peak_memory(self):
        """Get peak memory for each device."""
        peak_mem_list = []

        for rank_id in self._cluster_rank_ids:
            file_path = self._get_memory_file_for_each_device(self._target_dir_path, rank_id)
            file_content = self._get_file_content(file_path)
            capacity = file_content.get('capacity')
            peak_mem = file_content.get('peak_mem')

            mem_dict = {
                'rank_id': rank_id,
                'capacity': capacity,
                'peak_mem': peak_mem
            }
            peak_mem_list.append(mem_dict)

        return peak_mem_list

    def _get_memory_file_for_each_device(self, path, device_id):
        """Get memory file for each device."""
        filename = self._summary_filename.format(device_id)
        file_path = os.path.join(path, filename)
        validate_and_normalize_path(
            file_path, raise_key='Invalid memory usage file path.'
        )

        return file_path

    @staticmethod
    def _get_file_content(file_path):
        """Get file content."""
        try:
            with open(file_path, 'r') as f_obj:
                file_content = json.load(f_obj)
        except (IOError, OSError, json.JSONDecodeError) as err:
            log.error('Error occurred when read memory file: %s', err)
            raise ProfilerIOException()

        return file_content


class ClusterFlopsAnalyser(ClusterAnalyser):
    """The analyser for analyzing the cluster flops."""
    _summary_filename = 'flops_summary_{}.json'

    def __init__(self, cluster_profiler_dir, device_id='0'):
        super().__init__(cluster_profiler_dir, device_id)

    def get_flops(self):
        """Get flops for each device."""
        flops_info_list = []
        max_flops = 0

        for rank_id in self._cluster_rank_ids:
            file_path = self._get_flops_file_for_each_device(self._target_dir_path, rank_id)

            if not os.path.exists(file_path):
                msg = 'profiler does not find file: {}'.format(file_path)
                log.warning(msg)
                continue

            file_content = self._get_file_content(file_path)

            if file_content.get('FLOPs'):
                max_flops = max(max_flops, file_content.get('FLOPs'))
            else:
                max_flops = max(max_flops, file_content.get('cube_FLOPs') + file_content.get('vec_FLOPs'))
            flops_dict = {
                'rank_id': rank_id
            }
            flops_dict.update(file_content)
            flops_info_list.append(flops_dict)

        # Normalize the flops by divide the max flops in all device.
        for flops_info in flops_info_list:
            if flops_info.get('FLOPs'):
                flops_info['FLOPs_norm'] = flops_info['FLOPs'] / max_flops
            else:
                flops_info['FLOPs_norm'] = (flops_info['cube_FLOPs'] + flops_info['vec_FLOPs']) / max_flops
                flops_info['FLOPs'] = flops_info['cube_FLOPs'] + flops_info['vec_FLOPs']

        return flops_info_list

    def _get_flops_file_for_each_device(self, path, device_id):
        """Get memory file for each device."""
        filename = self._summary_filename.format(device_id)
        file_path = os.path.join(path, filename)
        validate_and_normalize_path(
            file_path, raise_key='Invalid flops file path.'
        )

        return file_path

    @staticmethod
    def _get_file_content(file_path):
        """Get file content."""
        try:
            with open(file_path, 'r') as f_obj:
                file_content = json.load(f_obj)
        except (IOError, OSError, json.JSONDecodeError) as err:
            log.error('Error occurred when read flops file: %s', err)
            raise ProfilerIOException()

        return file_content


class ClusterHcclAnalyser(ClusterAnalyser):
    """The analyser for analyzing the cluster communication info."""
    _col_names = ['step_num', 'communication_cost', 'wait_cost']
    _hccl_filename = 'hccl_raw_{}.csv'

    def __init__(self, cluster_profiler_dir, device_id):
        super().__init__(cluster_profiler_dir, device_id)
        self._none_sort_col_names = []
        self._total_step_num = self._get_total_step_num()

    def get_cluster_link_info(self, condition=None):
        """Get cluster link info."""
        self._col_names = ["src_dst", "link_type", "communication_cost", "communication_size", "band_width"]
        if condition is None:
            condition = {}
        filter_condition = condition.get('filter_condition', {})
        sort_condition = condition.get('sort_condition')
        group_condition = condition.get('group_condition')
        self._analyser_cluster_link_info(filter_condition)
        self._filter_cluster_link_info(filter_condition)
        if sort_condition:
            self._sort(sort_condition)
        if group_condition:
            self._group(group_condition)

        return {
            'cluster_link_info': self._result,
            'size': self._cluster_link_info_size
        }

    def get_communication_data(self):
        """Get communication data for node-link graph."""
        device_list, _, _ = analyse_device_list_from_profiler_dir(self._profiling_dir)
        data = {}
        for device in device_list:
            hccl_filename = self._hccl_filename.format(device)
            file_path = os.path.join(self._profiling_dir, hccl_filename)
            file_path = validate_and_normalize_path(
                file_path, raise_key='Invalid memory usage file path.'
            )
            if not os.path.exists(file_path):
                log.error('Did not find the file: %s', file_path)
                raise ProfilerFileNotFoundException(msg='Did not find the file:{}'.format(file_path))
            df = pd.read_csv(file_path)
            cur_data = list(json.loads(df.to_json(orient="index")).values())
            for dt in cur_data:
                dt['link_info'] = json.loads(dt['link_info'])
                dt['communication_operator_cost'] = json.loads(dt['communication_operator_cost'])
            device_entry = "device" + device
            data[device_entry] = {}
            data[device_entry] = cur_data
        return data

    def _get_total_step_num(self):
        """Get the num of train step."""
        total_step_num = 0
        # Take the data of one of the device to get the total number of steps.

        entries = os.scandir(self._target_dir_path)
        for entry in entries:
            if entry.is_symlink():
                continue
            if entry.is_file() and entry.name.startswith('hccl_raw'):
                file_path = os.path.join(self._target_dir_path, entry.name)
                with open(file_path, 'r') as src_file:
                    lines = src_file.readlines()
                # The first row is col_name, the last row is the average.
                if len(lines) > 2:
                    total_step_num = len(lines) - 2
                break
        return total_step_num

    def _filter_cluster_link_info(self, filter_condition):
        """Filter cluster link info."""
        src_rank = filter_condition.get("src_rank", ".*")
        dst_rank = filter_condition.get("dst_rank", ".*")
        src_dst = src_rank + '-' + dst_rank
        src_dst_regex = '^' + src_dst + '$'
        link_type = filter_condition.get("link_type", ".")
        if src_rank == ".*" and dst_rank == ".*" and link_type == ".":
            self._cluster_link_info_size = len(self._result)
            return

        def inner_filter(item: list):
            # index0:src_dst_rank_id, index1:link type
            src_dst_pattern = re.match(src_dst_regex, item[0])
            link_type_pattern = re.match(link_type, item[1])
            if src_dst_pattern and link_type_pattern:
                return True
            return False

        self._result = list(filter(inner_filter, self._result))
        self._cluster_link_info_size = len(self._result)

    def _analyser_cluster_link_info(self, filter_condition):
        """Get cluster link info."""
        link_info = []
        step_id = filter_condition.get("step_id", 0)
        cluster_communication_info = self._get_cluster_communication_info(step_id)
        # index3:link_info
        cluster_link_info = [i[3] for i in cluster_communication_info]
        for item in cluster_link_info:
            for src_dst_key, src_dst_value in item.items():
                for link_type_key, link_type_value in src_dst_value.items():
                    # index0:communication_cost,index1:communication_size,communication_brand_width
                    link_info.append([src_dst_key, link_type_key, link_type_value[0],
                                      link_type_value[1], link_type_value[2]])

        self._result = link_info

    def _get_cluster_communication_info(self, step_num):
        """Get cluster communication info."""
        cluster_communication_info = list()
        for item in self._cluster_rank_ids:
            communication_info = self._get_communication_info(item, step_num)
            communication_info.append(item)
            cluster_communication_info.append(communication_info)
        self._cluster_communication_info_size = len(cluster_communication_info)

        return cluster_communication_info

    def _get_communication_info(self, device_id, step_num):
        """Get step trace info."""
        file_name = 'hccl_raw_{}.csv'.format(device_id)
        communication_file_path = \
            os.path.join(self._target_dir_path, file_name)
        communication_file_path = validate_and_normalize_path(
            communication_file_path, raise_key="Invalid  communication file path.")
        if not os.path.exists(communication_file_path):
            log.error('Did not find the file: %s', communication_file_path)
            raise ProfilerFileNotFoundException(msg='Did not find the file:{}'.format(communication_file_path))
        communication_info = list()
        step_num = str(step_num)
        with open(communication_file_path, 'r') as src_file:
            csv_reader = csv.reader(src_file)
            # when the step_num value is 0, it means the average value.
            # The last line of the step_trace_raw_{}_detail_time.csv records the average value.
            # The first element of the last line is '-'.
            step_num = '-' if step_num == '0' else step_num
            for row in csv_reader:
                if row[0] == step_num:
                    communication_info = row
                    break
        # Convert string to floating point and dictionary
        if communication_info:
            communication_info[1] = float(communication_info[1])
            communication_info[2] = float(communication_info[2])
            communication_info[3] = json.loads(communication_info[3])
        # Judge whether the communication operator information is recorded.
        # If recorded, the length of communication_info is 5.
        if len(communication_info) > 4:
            communication_info[4] = json.loads(communication_info[4])

        return communication_info

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.

                - step_id (int): The selected step id.
        """

        step_id = filter_condition.get("step_id", 0)
        self._result = self._get_cluster_communication_info(step_id)

    def _organize_query_result(self):
        """
        Organize the query result.

        Returns:
            dict, the query result.
        """
        result = list()
        # item[0]:step_num, item[1]:communication_cost, item[2]:wait_cost, item[3]:link_info,
        # item[4]:communication_operator_cost, item[5]:rank_id
        for item in self._result:
            communication_info = dict()
            communication_info["communication_info"] = item[1:5]
            communication_info["rank_id"] = item[5]
            result.append(communication_info)

        return {
            'total_step_num': self._total_step_num,
            'communication': result,
            'size': self._cluster_communication_info_size
        }



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Constant module."""


from enum import Enum


class MindInsightModules(Enum):
    """
    Enum definition for MindInsight error types.

    Note:
        Each enum value, excluding GENERAL, has an Errors class name starting with the enum value
        in Camel-Case referring to specific module.
    """
    GENERAL = 0
    LINEAGEMGR = 2
    DATAVISUAL = 5
    PROFILERMGR = 6
    SCRIPTCONVERTER = 7
    WIZARD = 9
    OPTIMIZER = 10
    DEBUGGER = 11


class GeneralErrors(Enum):
    """Enum definition for general errors."""
    UNKNOWN_ERROR = 0
    PARAM_TYPE_ERROR = 1
    PARAM_VALUE_ERROR = 2
    PARAM_MISSING_ERROR = 3
    PATH_NOT_EXISTS_ERROR = 4
    FILE_SYSTEM_PERMISSION_ERROR = 8
    PORT_NOT_AVAILABLE_ERROR = 9
    URL_DECODE_ERROR = 10
    COMPUTING_RESOURCE_ERROR = 11
    MEMORY_ERROR = 12
    SETTING_VALUE_ERROR = 13


class ProfilerMgrErrors(Enum):
    """Enum definition for profiler errors."""


class LineageMgrErrors(Enum):
    """Enum definition for lineage errors."""


class DebuggerErrors(Enum):
    """Enum definition for debugger errors."""


class DataVisualErrors(Enum):
    """Enum definition for datavisual errors."""
    RESTFUL_API_NOT_EXIST = 1
    REQUEST_METHOD_NOT_ALLOWED = 2
    MAX_COUNT_EXCEEDED_ERROR = 3
    TRAIN_JOB_NOT_EXIST = 5
    SUMMARY_LOG_PATH_INVALID = 6
    SUMMARY_LOG_IS_LOADING = 7
    NODE_NOT_IN_GRAPH_ERROR = 9
    PATH_NOT_DIRECTORY_ERROR = 10
    PLUGIN_NOT_AVAILABLE = 11
    GRAPH_NOT_EXIST = 12
    IMAGE_NOT_EXIST = 13
    SCALAR_NOT_EXIST = 14
    HISTOGRAM_NOT_EXIST = 15
    TRAIN_JOB_DETAIL_NOT_IN_CACHE = 16
    QUERY_STRING_CONTAINS_NULL_BYTE = 17
    TENSOR_NOT_EXIST = 18
    MAX_RESPONSE_DATA_EXCEEDED_ERROR = 19
    STEP_TENSOR_DATA_NOT_IN_CACHE = 20
    CRC_LENGTH_FAILED = 21
    TENSOR_TOO_LARGE = 22
    LANDSCAPE_NOT_EXIST = 23


class WizardErrors(Enum):
    """Enum definition for mindwizard errors."""


class OptimizerErrors(Enum):
    """Enum definition for optimizer errors."""
    SAMPLES_NOT_ENOUGH = 1
    CORRELATION_NAN = 2
    HYPER_CONFIG_ERROR = 3
    OPTIMIZER_TERMINATE = 4
    CONFIG_PARAM_ERROR = 5
    HYPER_CONFIG_ENV_ERROR = 6


class GraphDomainErrors(Enum):
    """Enum definition for graph domain errors."""
    UNKNOWN_DATA_TYPE_ERROR = 1
    TUPLE_GETITEM_INDEX_ERROR = 2
    UNKNOWN_TENSOR_ERROR = 3



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Explainer evaluation encapsulator."""

import copy

from mindinsight.explainer.encapsulator.explain_data_encap import ExplainDataEncap
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError


class EvaluationEncap(ExplainDataEncap):
    """Explainer evaluation encapsulator."""

    def query_explainer_scores(self, train_id):
        """Query evaluation scores."""
        job = self.job_manager.get_job(train_id)
        if job is None:
            raise TrainJobNotExistError(train_id)
        return copy.deepcopy(job.explainer_scores)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""GP Tuner."""
import warnings
import numpy as np

from scipy.stats import norm
from scipy.optimize import minimize
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Matern

from mindinsight.optimizer.common.enums import AcquisitionFunctionEnum, HyperParamKey
from mindinsight.optimizer.common.log import logger
from mindinsight.optimizer.tuners.base_tuner import BaseTuner
from mindinsight.optimizer.utils.param_handler import generate_arrays, match_value_type
from mindinsight.optimizer.utils.transformer import Transformer
from mindinsight.utils.exceptions import ParamValueError


class AcquisitionFunction:
    """
    It can be seen from the Gaussian process that the probability description of the objective
    function can be obtained by sampling. Sampling usually involves two aspects:
        - Explore: Explore new spaces, this sampling helps to estimate more accurate results;
        - Exploit: Sampling near the existing results (usually near the existing maximum value),
          hoping to find larger results.

    The purpose of the acquisition function is to balance these two sampling processes.
    Supported acquisition function:
        - Probability of improvement.
        - Expected improvement.
        - Upper confidence bound. The weighted sum of posterior mean and posterior standard deviation.
          formula: result = exploitation + βt * exploration, where βt are appropriate constants.

    Args:
        method (str): The method for acquisition function, including 'ucb', 'pi', and 'ei'.
        beta (float): trade-off param for upper confidence bound function.
        beta_decay (float): the decay for beta. Formula: beta = beta * beta_decay.
        beta_decay_delay (int): if the counter is bigger than beta_decay_delay, the beta begins to decay.
        xi (float): trade-off for expected improvement and probability of improvement.
    """
    def __init__(self, method: str, beta, xi, beta_decay=1, beta_decay_delay=0):
        self._beta = beta
        self._beta_decay = beta_decay
        self._beta_decay_delay = beta_decay_delay
        self._xi = xi
        self._method = method.lower()
        if self._method not in AcquisitionFunctionEnum.list_members():
            raise ParamValueError(error_detail="The 'method' should be in %s." % AcquisitionFunctionEnum.list_members())

        self._counter = 0

    def update(self):
        """Update k."""
        self._counter += 1

        if self._counter > self._beta_decay_delay and self._beta_decay < 1:
            self._beta *= self._beta_decay

    def ac(self, x, gp, y_max):
        """Acquisition Function."""
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mean, std = gp.predict(x, return_std=True)

        if self._method == AcquisitionFunctionEnum.UCB.value:
            # Upper confidence bound.
            res = mean + self._beta * std
        elif self._method == AcquisitionFunctionEnum.EI.value:
            # Expected improvement.
            u_f = (mean - y_max - self._xi)
            z = u_f / std
            res = u_f * norm.cdf(z) + std * norm.pdf(z)
        else:
            # Probability of improvement.
            z = (mean - y_max - self._xi) / std
            res = norm.cdf(z)

        return res


class GPBaseTuner(BaseTuner):
    """
    Tuner using gaussian process regressor.

    Args:
        method (str): The method for acquisition function, including 'ucb', 'pi', and 'ei'.
            Detail at AcquisitionFunction.
        beta (float): β, trade-off param for upper confidence bound function.
        beta_decay (float): the decay for beta. beta = beta * beta_decay.
        beta_decay_delay (int): if counter is bigger than beta_decay_delay, the beta begins to decay.
        xi (float): ξ, trade-off for expected improvement and probability of improvement.
        random_state (np.random.RandomState): if it is None, it will be assigned as RandomState.
    """
    def __init__(self,
                 method=AcquisitionFunctionEnum.UCB.value,
                 beta=2.576,
                 beta_decay=1,
                 beta_decay_delay=0,
                 xi=0.0,
                 random_state=None):
        self._random_state = self._get_random_state(random_state)
        self._utility_function = AcquisitionFunction(method=method,
                                                     beta=beta,
                                                     xi=xi,
                                                     beta_decay=beta_decay,
                                                     beta_decay_delay=beta_decay_delay)
        self._gp = GaussianProcessRegressor(
            kernel=Matern(nu=2.5),
            alpha=1e-6,
            normalize_y=True,
            n_restarts_optimizer=5,
            random_state=self._random_state
        )

    def _get_random_state(self, random_state=None):
        """Get random state."""
        if random_state is not None and not isinstance(random_state, (int, np.random.RandomState)):
            raise ParamValueError("The 'random_state' should be None, integer or np.random.RandomState.")
        if not isinstance(random_state, np.random.RandomState):
            random_state = np.random.RandomState(random_state)
        return random_state

    def _acq_max(self, gp, y_max, bounds, params_info, n_warmup=10000, n_iter=10):
        """Get max try calculated by acquisition function."""
        x_tries = generate_arrays(params_info, n_warmup)
        ys = self._utility_function.ac(x_tries, gp=gp, y_max=y_max)
        x_max = x_tries[ys.argmax()]
        max_acq = ys.max()

        x_seeds = generate_arrays(params_info, n_iter)
        for x_try in x_seeds:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = minimize(lambda x: -self._utility_function.ac(x.reshape(1, -1), gp=gp, y_max=y_max),
                               x_try.reshape(1, -1), bounds=bounds, method="L-BFGS-B")

            if not res.success:
                continue

            if isinstance(res.fun, float):
                res_fun = -res.fun
            else:
                res_fun = -res.fun[0]

            if max_acq is None or res_fun >= max_acq:
                x_max = match_value_type(x_max, params_info)
                max_acq = res_fun

        return np.clip(x_max, bounds[:, 0], bounds[:, 1])

    def suggest(self, params, target, params_info: dict):
        """Get suggest values."""
        bounds = []
        for param_info in params_info.values():
            bound = param_info[HyperParamKey.BOUND.value] if HyperParamKey.BOUND.value in param_info \
                else param_info['choice']
            bounds.append([min(bound), max(bound)])
        bounds = np.array(bounds)

        min_lineage_rows = 2
        if not np.array(params).any() or params.shape[0] < min_lineage_rows:
            logger.info("Without valid histories or the rows of lineages < %s, "
                        "parameters will be recommended randomly.", min_lineage_rows)
            suggestion = generate_arrays(params_info)
        # in scikit-learn<0.24.2, if np.std(target) include 0, errors would occur.
        elif (np.std(target, axis=0) == 0).any():
            logger.warning("The std of target: %s include zero. To avoid errors, "
                           "parameters will be recommended randomly.", target)
            suggestion = generate_arrays(params_info)
        else:
            self._gp.fit(params, target)
            suggestion = self._acq_max(
                gp=self._gp,
                y_max=target.max(),
                bounds=bounds,
                params_info=params_info
            )

        suggestion, user_defined_info = Transformer.transform_list_to_dict(params_info, suggestion)
        return suggestion, user_defined_info



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Saliency map encapsulator."""

from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.explainer.common.enums import ExplanationKeys, ImageQueryTypes
from mindinsight.explainer.encapsulator.explain_data_encap import ExplanationEncap


class SaliencyEncap(ExplanationEncap):
    """Saliency map encapsulator."""

    def query_saliency_maps(self,
                            train_id,
                            labels,
                            explainers,
                            limit,
                            offset,
                            sorted_name,
                            sorted_type,
                            prediction_types=None):
        """
        Query saliency maps.

        Args:
            train_id (str): Job ID.
            labels (list[str]): Label filter.
            explainers (list[str]): Explainers of saliency maps to be shown.
            limit (int): Maximum number of items to be returned.
            offset (int): Page offset.
            sorted_name (str): Field to be sorted.
            sorted_type (str): Sorting order, 'ascending' or 'descending'.
            prediction_types (list[str]): Prediction types filter. Default: None.

        Returns:
            tuple[int, list[dict]], total number of samples after filtering and list of sample result.
        """
        job = self.job_manager.get_job(train_id)
        if job is None:
            raise TrainJobNotExistError(train_id)

        samples = self._query_samples(job, labels, sorted_name, sorted_type, prediction_types)

        sample_infos = []
        obj_offset = offset * limit
        count = len(samples)
        end = count
        if obj_offset + limit < end:
            end = obj_offset + limit
        for i in range(obj_offset, end):
            sample = samples[i]
            sample_infos.append(self._touch_sample(sample, job, explainers))

        return count, sample_infos

    def _touch_sample(self, sample, job, explainers):
        """
        Final edit on single sample info.

        Args:
            sample (dict): Sample info.
            job (ExplainJob): Explain job.
            explainers (list[str]): Explainer names.

        Returns:
            dict, the edited sample info.
        """
        original = ImageQueryTypes.ORIGINAL.value
        overlay = ImageQueryTypes.OVERLAY.value

        sample_cp = sample.copy()
        sample_cp["image"] = self._get_image_url(job.train_id, sample['image'], original)
        for inference in sample_cp["inferences"]:
            new_list = []
            for saliency_map in inference[ExplanationKeys.SALIENCY.value]:
                if explainers and saliency_map["explainer"] not in explainers:
                    continue
                saliency_map[overlay] = self._get_image_url(job.train_id, saliency_map[overlay], overlay)
                new_list.append(saliency_map)
            inference[ExplanationKeys.SALIENCY.value] = new_list
        return sample_cp



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This module is aimed to provide with GraphRuns."""
from collections import defaultdict
from dataclasses import field, dataclass, asdict
from typing import List

from mindinsight.debugger.common.exceptions.exceptions import DebuggerHistoryNotFoundError, \
    DebuggerHistoryValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import Streams, enter_tag


@dataclass
class GraphRun:
    """Graph run data class."""
    count: int = 0
    graph_name: str = 'None'
    has_data: bool = False
    sub_graph_names: List[str] = field(default_factory=list)

    @staticmethod
    def _get_graph_id(graph_name):
        """Get graph id by graph name."""
        graph_id_str = graph_name.rsplit('_', 1)[-1]
        graph_id = int(graph_id_str) if graph_id_str.isdigit() else 0
        return graph_id

    def __lt__(self, other):
        if self.count != other.count:
            return self.count < other.count
        cur_graph_id = self._get_graph_id(self.graph_name)
        other_graph_id = self._get_graph_id(other.graph_name)
        return cur_graph_id < other_graph_id


class GraphRunsOperator:
    """Manage Graph runs information."""

    def __init__(self, cache):
        self._cache = cache
        self._multi_card_graph_stream = cache.get_stream_handler(Streams.GRAPH)
        self._graph_history_stream = cache.get_stream_handler(Streams.GRAPH_HISTORY)

    def get_graph_runs(self, rank_id):
        """
        Get graph runs info of specified rank.

        Args:
            rank_id (int): The rank id.

        Returns:
            dict, the graph runs. The format is like {'graph_runs': List[GraphRun]}
            The GraphRun object = {
                "count": int,
                "graph_name": str,
                "has_data": bool,
                "sub_graph_names": List[str]
            }
        """
        graph_history = self._graph_history_stream.get(rank_id)
        graph_run_count = graph_history.get_total_count()
        if not graph_run_count:
            log.warning("No graph runs found.")
            return {'graph_runs': []}
        graph_run_map = self.get_basic_graph_runs(rank_id)
        graph_run_map = self.update_has_data_field(graph_run_map, rank_id)
        res = self.get_sorted_graph_runs(graph_run_map)
        return {'graph_runs': res}

    @enter_tag
    def get_basic_graph_runs(self, rank_id):
        """
        Get basic graph runs.

        Args:
            rank_id (int): The rank id.

        Returns:
            dict, the key is count id, the value is dict of <root_graph_id, GraphRun>.
        """
        graph_history = self._graph_history_stream.get(rank_id)
        root_graphs = self._multi_card_graph_stream.get_graph_handler_by_rank_id(rank_id).root_graphs
        graph_run_map = defaultdict(dict)
        for root_graph_id, graph_run_ids in graph_history.history.items():
            root_graph = root_graphs.get(root_graph_id)
            if root_graph is None:
                log.error("Didn't find relative graph info of graph %s. "
                          "Please check the dump structure.", root_graph_id)
                raise DebuggerHistoryNotFoundError(root_graph_id)
            for graph_run_id in graph_run_ids:
                graph_run = GraphRun(count=graph_run_id + 1,
                                     graph_name=root_graph.graph_name,
                                     sub_graph_names=root_graph.sub_graph_names)
                graph_run_map[graph_run_id][root_graph_id] = graph_run
        return graph_run_map

    def update_has_data_field(self, graph_run_map, rank_id):
        """Update has_data field in GraphRun."""
        dumped_steps = self._graph_history_stream.get_dumped_step(rank_id)
        for root_graph_id, dumped_steps in dumped_steps.items():
            for step_id in dumped_steps:
                graph_run = graph_run_map.get(step_id, {}).get(root_graph_id)
                if graph_run is None:
                    log.error("Didn't find step_id %s in graph history of rank %s", step_id, rank_id)
                    raise DebuggerHistoryValueError(root_graph_id)
                graph_run.has_data = True
        return graph_run_map

    @staticmethod
    def get_sorted_graph_runs(graph_run_map):
        """Turn graph run map into ordered list."""
        graph_runs = []
        for graph_runs_per_step in graph_run_map.values():
            for graph_run in graph_runs_per_step.values():
                graph_runs.append(graph_run)
        graph_runs.sort()
        graph_runs = [asdict(graph_run) for graph_run in graph_runs]
        return graph_runs



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Command module."""
import os
import sys
import argparse

import mindinsight
from mindinsight.wizard.create_project import CreateProject


def cli_entry():
    """Entry point for mindwizard CLI."""

    permissions = os.R_OK | os.W_OK | os.X_OK
    os.umask(permissions << 3 | permissions)

    parser = argparse.ArgumentParser(
        prog='mindwizard',
        description='MindWizard CLI entry point (version: {})'.format(mindinsight.__version__))

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ({})'.format(mindinsight.__version__))
    command = CreateProject()
    command.add_arguments(parser)
    argv = sys.argv[1:]
    if not argv or argv[0] == 'help':
        argv = ['-h']

    args = parser.parse_args(argv)
    command.invoke(vars(args))


if __name__ == '__main__':
    cli_entry()



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Start mindinsight service."""

import argparse
import os
import re
import sys
import socket
from importlib import import_module

import psutil

from mindinsight.conf import settings, get_web_address
from mindinsight.utils.command import BaseCommand
from mindinsight.utils.exceptions import PortNotAvailableError, SettingValueError
from mindinsight.modelarts.exceptions import PortReuseException
from mindinsight.utils.hook import HookUtils
from mindinsight.utils.hook import init

MIN_SESSION_NUM = 1
MAX_SESSION_NUM = 2
# The unit is MB
MIN_MEM_LIMIT_VALUE = 6 * 1024
# The limit for int32.
MAX_MEM_LIMIT_VALUE = 2147483647


class WorkspaceAction(argparse.Action):
    """Workspace action class definition."""

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        workspace = os.path.realpath(values)
        setattr(namespace, self.dest, workspace)


class PortAction(argparse.Action):
    """Port action class definition."""

    MIN_PORT = 1
    MAX_PORT = 65535

    OPEN_PORT_LIMIT = 1024

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        port = values
        if not self.MIN_PORT <= port <= self.MAX_PORT:
            parser.error(f'{option_string} should be chosen from {self.MIN_PORT} to {self.MAX_PORT}')

        setattr(namespace, self.dest, port)


class UrlPathPrefixAction(argparse.Action):
    """Url Path prefix action class definition."""

    INVALID_SEGMENTS = ('.', '..')
    REGEX = r'^[a-zA-Z0-9_\-\.]+$'

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Inherited __call__ method from argparse.Action.

        Args:
            parser (ArgumentParser): Passed-in argument parser.
            namespace (Namespace): Namespace object to hold arguments.
            values (object): Argument values with type depending on argument definition.
            option_string (str): Optional string for specific argument name. Default: None.
        """
        prefix = values
        segments = prefix.split('/')
        for index, segment in enumerate(segments):
            if not segment and index in (0, len(segments) - 1):
                continue
            if segment in self.INVALID_SEGMENTS or not re.match(self.REGEX, segment):
                parser.error(f'{option_string} value is invalid url path prefix')

        setattr(namespace, self.dest, prefix)


class Command(BaseCommand):
    """
    Start mindinsight service.
    """

    name = 'start'
    description = 'startup mindinsight service'

    def add_arguments(self, parser):
        """
        Add arguments to parser.

        Args:
            parser (ArgumentParser): Specify parser to which arguments are added.
        """
        parser.add_argument(
            '--workspace',
            type=str,
            action=WorkspaceAction,
            help="""
                Specify path for user workspace. Default is
                $HOME/mindinsight.
            """)

        parser.add_argument(
            '--port',
            type=int,
            action=PortAction,
            help="""
                Custom port ranging from %s to %s. Default value is %s.
            """ % (PortAction.MIN_PORT, PortAction.MAX_PORT, settings.PORT))

        parser.add_argument(
            '--url-path-prefix',
            type=str,
            action=UrlPathPrefixAction,
            help="""
                Custom URL path prefix for web page address. URL path prefix 
                consists of segments separated by slashes. Each segment supports
                alphabets / digits / underscores / dashes / dots, but not single 
                dot or double dots. Default value is ''.
            """)

        for hook in HookUtils.instance().hooks():
            hook.register_startup_arguments(parser)

    def update_settings(self, args):
        """
        Update settings.

        Args:
            args (Namespace): parsed arguments to hold customized parameters.
        """
        kwargs = {}
        for key, value in args.__dict__.items():
            if value is not None:
                kwargs[key] = value

        init(**kwargs)

    def run(self, args):
        """
        Execute for start command.

        Args:
            args (Namespace): Parsed arguments to hold customized parameters.
        """
        for key, value in args.__dict__.items():
            if value is not None:
                self.logfile.info('%s = %s', key, value)

        try:
            self.check_port()
            self.check_debugger_port()
            self.check_offline_debugger_setting()
        except (PortNotAvailableError, SettingValueError) as error:
            self.console.error(error.message)
            self.logfile.error(error.message)
            sys.exit(1)
        except PortReuseException as ex:
            self.console.error(ex.message)
            self.logfile.error(ex.message)
            sys.exit(2)

        self.console.info('Workspace: %s', os.path.realpath(settings.WORKSPACE))
        self.console.info('Summary base dir: %s', os.path.realpath(settings.SUMMARY_BASE_DIR))

        run_module = import_module('mindinsight.backend.run')
        state_result = run_module.start()

        self.logfile.info('Start mindinsight done. Workspace is %s,', settings.WORKSPACE)
        self.logfile.info('Port is %s,', settings.PORT)
        self.logfile.info('Summary base dir is %s', settings.SUMMARY_BASE_DIR)
        return state_result

    def check_port(self):
        """Check port."""
        if os.getuid() != 0 and settings.PORT < PortAction.OPEN_PORT_LIMIT:
            raise PortNotAvailableError(
                f'Port {settings.PORT} below {PortAction.OPEN_PORT_LIMIT} is not allowed by current user.')

        connections = psutil.net_connections()
        for connection in connections:
            if connection.status != 'LISTEN':
                continue
            if connection.laddr.port == settings.PORT:
                last_log_base_dir = self._get_log_base_dir(settings.PORT)
                web_address = get_web_address()
                if last_log_base_dir == settings.SUMMARY_BASE_DIR:
                    raise PortReuseException(
                        f'Reusing MindInsight on port {settings.PORT}. (pid {connection.pid}, ' \
                        f'Use "mindinsight stop --port {settings.PORT}" to stop it.)\n' \
                        f'{web_address}')
                raise PortNotAvailableError(
                    f'MindInsight could not bind to port {settings.PORT}, ' \
                    f'it was already in use (pid {connection.pid})')

    def _get_log_base_dir(self, port):
        """Get log base dir."""
        log_base_dir = ''
        log_path = os.path.join(settings.WORKSPACE, 'log/scripts/start.{}.log'.format(port))
        log_key_word = 'Summary base dir is '
        try:
            with open(log_path) as f:
                logs = f.readlines()
                for line in reversed(logs):
                    if log_key_word in line:
                        log_base_dir = line.split(log_key_word)[1].strip()
                        break
        except (FileNotFoundError, PermissionError) as error:
            log.info("Get log base dir error.")
            log.info(str(error))

        return log_base_dir

    def check_debugger_port(self):
        """Check if the debugger_port is available"""
        if not settings.ENABLE_DEBUGGER:
            return
        ip = settings.HOST
        debugger_port = settings.DEBUGGER_PORT
        port = settings.PORT
        if debugger_port == port:
            raise PortNotAvailableError("The ports for debugger and web page are both %s,"
                                        "please use another port as debugger-port." % debugger_port)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((ip, debugger_port))
            s.shutdown(2)
            raise PortNotAvailableError(f'Debugger-port {ip}:{debugger_port} is not available for MindInsight')
        except socket.error:
            return

    def check_offline_debugger_setting(self):
        """Check if the offline debugger setting is legal"""
        mem_limit = settings.OFFLINE_DEBUGGER_MEM_LIMIT
        session_num = settings.MAX_OFFLINE_DEBUGGER_SESSION_NUM
        if not isinstance(mem_limit, int) or isinstance(mem_limit, bool) or \
                mem_limit < MIN_MEM_LIMIT_VALUE or mem_limit > MAX_MEM_LIMIT_VALUE:
            raise SettingValueError("Offline debugger memory limit should be integer ranging from {} to {} MB, but got"
                                    " {}. Please check the environment variable MINDINSIGHT_OFFLINE_DEBUGGER_MEM_LIMIT"
                                    .format(MIN_MEM_LIMIT_VALUE, MAX_MEM_LIMIT_VALUE, mem_limit))
        if not isinstance(session_num, int) or isinstance(session_num, bool) or \
                session_num < MIN_SESSION_NUM or session_num > MAX_SESSION_NUM:
            raise SettingValueError("Max offline debugger session number should be integer ranging from {} to {}, but "
                                    "got {}. Please check the environment variable "
                                    "MINDINSIGHT_MAX_OFFLINE_DEBUGGER_SESSION_NUM"
                                    .format(MIN_SESSION_NUM, MAX_SESSION_NUM, session_num))



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define the watchpoint stream handler."""
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError, \
    DebuggerParamTypeError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.conditionmgr.condition import ParamTypeEnum
from mindinsight.debugger.conditionmgr.condition import ValueTypeEnum
from mindinsight.debugger.proto.debug_grpc_pb2 import SetCMD
from mindinsight.debugger.stream_cache.watchpoint import Watchpoint, WatchpointHit, \
    WatchNodeTree
from mindinsight.debugger.stream_handler.base_handler import StreamHandlerBase

RANGE_START = 'range_start_inclusive'
RANGE_END = 'range_end_inclusive'


class WatchpointHandler(StreamHandlerBase):
    """Watchpoint Handler."""

    def __init__(self):
        self._watchpoints = {}
        # list of ids of new created watchpoints
        self._created_watchpoints = []
        # list of SetCMD of watchpoints to be deleted
        self._deleted_watchpoints = []
        # dict of <id, Watchpoint> of watchpoints to be updated
        self._updated_watchpoints = {}
        # the collection of watched node full names, which have been sent to MindSpore
        self._latest_id = 0
        self._cache_set_cmd = {}
        # whether the watchpoint list has been changed since last step
        self._outdated = False

    @property
    def empty(self):
        """Check the there is no watchpoint."""
        return not self._watchpoints

    def set_outdated(self):
        """"Set outdated as True."""
        self._outdated = True

    def put(self, value):
        """
        Put Watchpoint into watchpoint handler.

        Args:
            value (Watchpoint): The name of nodes that have been chosen.
        """
        new_id = value.watchpoint_id
        self._watchpoints[new_id] = value
        self._created_watchpoints.append(new_id)
        self._updated_watchpoints[new_id] = value
        self._latest_id = new_id
        log.debug("Put watchpoint %d into cache.", new_id)

    def sync_set_cmd(self, set_cmds):
        """Clean temp watchpoints."""
        self._outdated = False
        self._created_watchpoints = []
        self._deleted_watchpoints = []
        self._updated_watchpoints = {}
        for set_cmd in set_cmds:
            self._cache_set_cmd[set_cmd.id] = set_cmd

    def clean_cache_set_cmd(self, set_cmd):
        """Clean cache set command."""
        self._cache_set_cmd.pop(set_cmd.id, None)

    def get_watchpoint_by_id(self, watchpoint_id):
        """Get watchpoint by watchpoint id."""
        res = self.get(watchpoint_id)
        watchpoint = res.get('watch_points')[0]

        return watchpoint

    def get(self, filter_condition=None):
        """
        Get the watchpoints.

        Args:
            filter_condition (Union[None, int]): The filter conditions. Get watchpoint by
                id. If None, return all watchpoint. Default: None.

        Returns:
            dict, the watchpoint list.
        """
        reply = []
        if not filter_condition:
            # get watch condition list
            for _, watchpoint in self._watchpoints.items():
                watchpoint_info = watchpoint.get_watch_condition_info()
                reply.append(watchpoint_info)
        else:
            self.validate_watchpoint_id(filter_condition)
            reply = [self._watchpoints.get(filter_condition)]

        log.debug("get the watch points with filter_condition:%s", filter_condition)

        return {'watch_points': reply}

    def get_pending_commands(self):
        """
        Get all watchpoint in SetCMD proto format.

        Returns:
            list[SetCMD], updated watchpoint to be sent to MindSpore.
        """
        newly_set_cmds = []
        for _, watchpoint in self._updated_watchpoints.items():
            set_cmd = SetCMD()
            set_cmd.id = watchpoint.watchpoint_id
            set_cmd.delete = False
            newly_set_cmds.append(set_cmd)
        newly_set_cmds.extend(self._deleted_watchpoints)
        self.sync_set_cmd(newly_set_cmds)

        return list(self._cache_set_cmd.values())

    def is_recheckable(self):
        """
        Check if current status is able to recheck.

        Returns:
            bool, if enable to recheck.
        """
        return self._outdated

    def set_watch_nodes(self, graph, graph_stream, watch_point_id, graph_name=None, rank_id=0):
        """
        set watch nodes for graph.

        Args:
            graph (dict): The graph with list of nodes.
            graph_stream (GraphHandler): The graph handler.
            watch_point_id (int): The id of watchpoint.
            graph_name (str): The graph name.
            rank_id (int): The rank id.
        """
        if not (watch_point_id and graph):
            return
        log.debug("add watch flags")
        watchpoint = self._watchpoints.get(watch_point_id)
        self._set_watch_status_recursively(graph, graph_stream, watchpoint, graph_name, rank_id)

    def _set_watch_status_recursively(self, graph, graph_stream, watchpoint, graph_name=None, rank_id=0):
        """Set watch status to graph."""
        if graph.get('children'):
            self._set_watch_status_recursively(
                graph.get('children'), graph_stream, watchpoint, graph_name, rank_id=0)

        if graph.get('nodes'):
            _ = self._set_watch_state_for_nodes(graph['nodes'], graph_stream, watchpoint, graph_name, rank_id)

    def _set_watch_state_for_nodes(self, nodes, graph_stream, watchpoint, graph_name, rank_id=0):
        """
        Set watch state for nodes.

        Args:
            nodes (list[Node]): List of node info.

        Returns:
            int, the number of all watched nodes.
        """
        all_watched_num = 0
        valid_node_num = len(nodes)
        # initialize the state of current node.
        state = WatchNodeTree.NOT_WATCH
        for node in nodes:
            node_name = node.get('name')
            # search result could have `nodes` in nodes object

            disable = True
            if node.get('nodes'):
                flag = self._set_watch_state_for_nodes(node.get('nodes'), graph_stream, watchpoint, graph_name, rank_id)
            else:
                full_name = graph_stream.get_full_name(node_name, graph_name)
                new_node_name = node_name if graph_name is None else '/'.join([graph_name, node_name])
                flag = watchpoint.get_node_status(new_node_name, node.get('type'), full_name, rank_id)
            node['watched'] = flag
            if flag == WatchNodeTree.NOT_WATCH:
                node['disable'] = disable
                continue
            state = WatchNodeTree.PARTIAL_WATCH
            if flag == WatchNodeTree.INVALID:
                valid_node_num -= 1
            elif flag == WatchNodeTree.TOTAL_WATCH:
                disable = False
                all_watched_num += 1
            node['disable'] = disable
        # update the watch status of current node
        if not valid_node_num:
            state = WatchNodeTree.INVALID
        elif all_watched_num == valid_node_num:
            state = WatchNodeTree.TOTAL_WATCH
        return state

    def create_watchpoint(self, condition_mgr, watch_condition, watch_nodes=None, watch_point_id=None, name=None,
                          device_amount=8):
        """
        Create watchpoint.
        Args:
            condition_mgr (ConditionMgr): Instance of ConditionMgr.
            watch_condition (dict): The watch condition.
                "condition": {
                    id: "tensor_too_large",
                    "params": [
                        {
                            "name": "abs_mean_gt",
                            "value": 1.1
                        }
                    ]
                }

                - id (str): Id of condition.
                - param (list[dict]): The list of param for this condition.
            watch_nodes (dict[list[NodeBasicInfo]]): The list of node basic info.
            watch_point_id (int): The id of watchpoint.
            name (str): The name of watchpoint.
            device_amount (int): The amount of devices.

        Returns:
            int, the new id of watchpoint.
        """
        validate_watch_condition(condition_mgr, watch_condition)
        watch_condition = set_default_param(condition_mgr, watch_condition)
        new_id = self._latest_id + 1
        watchpoint = Watchpoint(new_id, watch_condition, name)
        if watch_nodes:
            for rank_id, watch_nodes_for_device in watch_nodes.items():
                validate_rank_id(rank_id, device_amount)
                watchpoint.add_nodes(watch_nodes_for_device, rank_id)
        elif watch_point_id:
            self.validate_watchpoint_id(watch_point_id)
            watchpoint.copy_nodes_from(self._watchpoints.get(watch_point_id))
        self.put(watchpoint)
        self._outdated = True
        return new_id

    def update_watchpoint(self, watch_point_id, watch_nodes, watched=False, rank_id=0):
        """
        Update watchpoint.

        Args:
            watch_point_id (int): The id of watchpoint.
            watch_nodes (list[NodeBasicInfo]): The list of node basic info.
            watched (bool): The update operator on nodes. If False, remove nodes from watch nodes.
                If True, add nodes to watch nodes. Default: False.
            rank_id (int): The rank id.
        """
        self.validate_watchpoint_id(watch_point_id)
        watchpoint = self._watchpoints.get(watch_point_id)
        if watched:
            watchpoint.add_nodes(watch_nodes, rank_id)
        else:
            watchpoint.remove_nodes(watch_nodes, rank_id)
        self._updated_watchpoints[watch_point_id] = watchpoint
        self._outdated = True
        log.debug("Update watchpoint %d in cache.", watch_point_id)

    def delete_watchpoint(self, watch_point_id=None):
        """
        Delete watchpoint.

        Args:
            watch_point_id (Union[None, int]): The id of watchpoint.
                If None, delete all watchpoints. Default: None.
        """
        if watch_point_id is None:
            watch_point_ids = [sub_id for sub_id, _ in self._watchpoints.items()]
        else:
            self.validate_watchpoint_id(watch_point_id)
            watch_point_ids = [watch_point_id]
        for single_id in watch_point_ids:
            self._delete_single_watchpoint(single_id)
        self._outdated = True

    def _delete_single_watchpoint(self, watch_point_id):
        """
        Delete single watchpoint.

        Args:
            watch_point_id (int): The id of watchpoint.
        """
        self._watchpoints.pop(watch_point_id)
        # if the watchpoint has not been created by MindSpore, clean the relative cache directly
        if watch_point_id in self._created_watchpoints:
            self._created_watchpoints.remove(watch_point_id)
            self._updated_watchpoints.pop(watch_point_id)
            log.debug("Cancel create watchpoint %d in cache.", watch_point_id)
            return
        set_cmd = SetCMD()
        set_cmd.id = watch_point_id
        set_cmd.delete = True
        self._deleted_watchpoints.append(set_cmd)
        log.debug("Delete watchpoint %d in cache.", watch_point_id)

    def validate_watchpoint_id(self, watch_point_id):
        """Validate watchpoint id."""
        if not isinstance(watch_point_id, int):
            log.error("Invalid watchpoint id %s. The watch point id should be int.", watch_point_id)
            raise DebuggerParamTypeError("Watchpoint id should be int type.")
        if watch_point_id and watch_point_id not in self._watchpoints:
            log.error("Invalid watchpoint id: %d.", watch_point_id)
            raise DebuggerParamValueError("Invalid watchpoint id: {}".format(watch_point_id))


class MultiCardWatchpointHitHandler:
    """Multi-card Watchpoint-hit Handler."""

    def __init__(self):
        self.watchpoint_hit_handlers = {0: WatchpointHitHandler()}

    def get_hit_handler_by_rank_id(self, rank_id=0):
        """Get handler by rank id."""
        if rank_id in self.watchpoint_hit_handlers:
            return self.watchpoint_hit_handlers.get(rank_id)
        log.error("There is no rank id %d.", rank_id)
        raise ValueError

    def put(self, value):
        """Put watchpoint hit into cache."""
        for rank_id, tensor_hit_values in value.items():
            if rank_id not in self.watchpoint_hit_handlers:
                self.watchpoint_hit_handlers[rank_id] = WatchpointHitHandler()
            cur_hit_handler = self.watchpoint_hit_handlers[rank_id]
            for tensor_hit_value in tensor_hit_values:
                cur_hit_handler.put(tensor_hit_value)

    def get(self, filter_condition=None, rank_id=0):
        """Get the graph of specific node for specific device."""
        if rank_id in self.watchpoint_hit_handlers:
            return self.watchpoint_hit_handlers.get(rank_id).get(filter_condition)
        log.error("There is no rank id %d.", rank_id)
        raise ValueError

    def update_tensor_history(self, tensor_history, rank_id):
        """
        Add hit flag to tensor history.

        Args:
            tensor_history (dict): The tensor history.
            rank_id (int): The rank id.
        """
        if rank_id in self.watchpoint_hit_handlers:
            self.watchpoint_hit_handlers[rank_id].update_tensor_history(tensor_history)
        else:
            for tensor_info in tensor_history.get('tensor_history'):
                tensor_info['is_hit'] = False

    def check_rank_id(self, rank_id):
        """check if has the rank id."""
        return rank_id in self.watchpoint_hit_handlers

    def clean(self):
        """Clean cache."""
        self.__init__()


class WatchpointHitHandler(StreamHandlerBase):
    """Watchpoint hit handler."""

    def __init__(self):
        # dict of <ui node_name, dict of <slot, WatchpointHit>>,
        self._ordered_hits = []
        # filtered results in self._ordered_hits
        self._filtered_hits = []
        self._multi_graph_hits = {}

    @property
    def empty(self):
        """Whether the watchpoint hit is empty."""
        return not self._multi_graph_hits

    def put(self, value):
        """
        Put value into watchpoint hit cache. Called by grpc server.

        Args:
            value (dict): The watchpoint hit info.

                - tensor_proto (TensorProto): The message about hit tensor.
                - watchpoint (Watchpoint): The Watchpoint that a node hit.
                - node_name (str): The UI node name.
                - graph_name (str): The graph name.
                - error_code (int): The code of errors.
        """
        watchpoint_hit = WatchpointHit(
            tensor_proto=value.get('tensor_proto'),
            watchpoint=value.get('watchpoint'),
            node_name=value.get('node_name'),
            graph_name=value.get('graph_name')
        )
        if 'error_code' in value.keys():
            watchpoint_hit.error_code = value.get('error_code')
        # get all hit watchpoints according to node name ans tensor slot
        watchpoint_hits = self._get_watchpoints_by_tensor_name(watchpoint_hit.graph_name, watchpoint_hit.node_name,
                                                               watchpoint_hit.slot)
        if watchpoint_hit not in watchpoint_hits:
            watchpoint_hits.append(watchpoint_hit)

    def _get_watchpoints_by_tensor_name(self, graph_name, node_name, slot):
        """
        Get hit tensors according to ui node name and slot.

        Args:
            node_name (str): The node name.
            slot (str): The tensor slot.

        Returns:
            list, list of watchpoints.
        """
        index = self._multi_graph_hits.get((graph_name, node_name))
        if index is None:
            hit_node = {}
            self._ordered_hits.append(hit_node)
            index = len(self._ordered_hits) - 1
            self._multi_graph_hits[(graph_name, node_name)] = index

        hit_node = self._ordered_hits[index]
        hit_tensors = hit_node.get(slot)
        if hit_tensors is None:
            hit_tensors = []
            hit_node[slot] = hit_tensors

        return hit_tensors

    def get(self, filter_condition=None):
        """
        Get watchpoint hit list.

        Args:
            filter_condition (str): Get the watchpoint hit according to specified node name.
                If not given, get all watchpoint hits. Default: None.

        Returns:
            dict, the watchpoint hit list.
        """
        reply = None
        if filter_condition is None:
            log.debug("Get all watchpoint hit list.")
            reply = self.get_watchpoint_hits()
        else:
            log.debug("Get the watchpoint for node: <%s>.", filter_condition)
            index = self._multi_graph_hits.get(("", filter_condition))
            if index is not None:
                reply = self._ordered_hits[index]
        return reply

    def group_by(self, group_condition):
        """
        Return the watchpoint hits by group condition.

        Args:
            group_condition (dict): The group conditions.

                - limit (int): The limit number of watchpoint hits each page.
                - offset (int): The page offset.
                - focused_node (dict): The focused node.
                  If the specified node is hit, return the page where the node is located.

                    - node_name (str): The retrieved node name.
                    - graph_name (str): The retrieved graph name.
                - graph_id (int): The graph id.
                - watchpoint_id (int): The watchpoint id.

        Returns:
            dict, the watchpoint hit list.
        """
        if not isinstance(group_condition, dict):
            log.error("The parameter <group_condition> should be dict")
            raise DebuggerParamTypeError("The parameter <group_condition> should be dict.")
        log.debug("Group watch_point_hits by group_condition: %s.", group_condition)
        limit = group_condition.get('limit', 0)
        if not isinstance(limit, int) or limit <= 0:
            log.error("Param 'limit' is invalid or absent. 'limit' should be a positive integer.")
            raise DebuggerParamValueError("Param 'limit' is invalid or absent. 'limit' should be a positive integer.")
        # Get all watch_point_hit info
        watch_point_hits = []
        for watch_point_hit in self._ordered_hits:
            self._get_tensors(watch_point_hit, watch_point_hits)

        graph_id = group_condition.get('graph_id')
        watchpoint_id = group_condition.get('watchpoint_id')

        # Filter by graph and watch_point
        if graph_id:
            log.debug("Filter by graph: %s", graph_id)
            watch_point_hits = self._filter_by_graph(graph_id, watch_point_hits)
        else:
            watch_point_hits.sort(key=lambda hit: int(hit['graph_id'].split("_")[-1]))
        if watchpoint_id:
            log.debug("Filter by watchpoint: %s.", watchpoint_id)
            watch_point_hits = self._filter_by_watchpoint(watchpoint_id, watch_point_hits)

        total = len(watch_point_hits)
        focused_node = group_condition.get('focused_node')
        if focused_node is None:
            watch_point_hits = self._get_by_offset(group_condition, watch_point_hits)
        else:
            watch_point_hits = self._get_by_name(group_condition, watch_point_hits)
        reply = {}
        if watch_point_hits:
            reply = {'offset': group_condition.get('offset', 0),
                     'total': total,
                     'watch_point_hits': watch_point_hits
                     }
        return reply

    def _filter_by_graph(self, graph_id, watch_point_hits):
        """Return the list of watchpoint hits filtered by graph_id."""
        if not isinstance(graph_id, str):
            log.error("Param graph_id is not a string.")
            raise DebuggerParamValueError("Param graph_id is not a string.")
        log.debug("Filter by graph: %s.", graph_id)
        filtered_watchpoint_hits = []
        log.debug("Before filtering, watch_point_hits are: %s.", watch_point_hits)
        for watch_point_hit in watch_point_hits:
            if watch_point_hit['graph_id'] == graph_id:
                filtered_watchpoint_hits.append(watch_point_hit)
        log.debug("Filtered watch_point_hits: %s", filtered_watchpoint_hits)
        return filtered_watchpoint_hits

    def _filter_by_watchpoint(self, watchpoint_id, watch_point_hits):
        """Return the list of watchpoint hits filtered by watchpoint_id"""
        if not isinstance(watchpoint_id, int):
            log.error("Param watchpoint_id is not an integer.")
            raise DebuggerParamValueError("Param watchpoint_id is not an integer.")
        log.debug("Filter by watchpoint: %s", watchpoint_id)
        log.debug("Before filtering, watch_point_hits are: %s.", watch_point_hits)
        filtered_watchpoint_hits = []
        for watch_point_hit in watch_point_hits:
            if watchpoint_id in watch_point_hit['watchpoint_id']:
                filtered_watchpoint_hits.append(watch_point_hit)

        log.debug("Filtered watch_point_hits: %s", filtered_watchpoint_hits)
        return filtered_watchpoint_hits

    def _get_by_offset(self, group_condition, watch_point_hits):
        """Return the list of watchpoint hits on the offset page."""
        limit = group_condition.get('limit', 10)
        offset = group_condition.get('offset', 0)
        if not isinstance(limit, int) or not isinstance(offset, int):
            log.error("Param limit or offset is not a integer")
            raise DebuggerParamValueError("Param limit or offset is not a integer")

        total = len(watch_point_hits)

        if limit * offset >= total and offset != 0:
            log.error("Param offset out of bounds")
            raise DebuggerParamValueError("Param offset out of bounds")

        if total == 0:
            return []

        hits_in_offset = []
        for hit in watch_point_hits[(limit * offset): (limit * (offset + 1))]:
            hits_in_offset.append(hit)

        return hits_in_offset

    def _get_by_name(self, group_condition, watch_point_hits):
        """Return the list of watchpoint hits by the group condition."""
        limit = group_condition.get('limit')
        focused_node = group_condition.get('focused_node')
        if not isinstance(focused_node, dict):
            log.error("Param focused_node is not a dict.")
            raise DebuggerParamValueError("Param focused_node is not a dict.")

        graph_name = focused_node.get('graph_name')
        node_name = focused_node.get('node_name')
        index = None
        for i, watch_point_hit in enumerate(watch_point_hits):
            if watch_point_hit['graph_name'] == graph_name and watch_point_hit['node_name'] == node_name:
                index = i
                break

        if index is not None:
            group_condition['offset'] = index//limit

        return self._get_by_offset(group_condition, watch_point_hits)

    def get_watchpoint_hits(self):
        """Return the list of watchpoint hits."""
        log.debug("Get watch_point_hits.")
        watch_point_hits = []
        for watchpoint_hits in self._ordered_hits:
            self._get_tensors(watchpoint_hits, watch_point_hits)
        log.debug("Watch_point_hits: %s.", watch_point_hits)

        return {'watch_point_hits': watch_point_hits}

    def _get_tensors(self, watchpoint_hits, watch_point_hits):
        """Get the tensors info for the watchpoint_hits."""
        log.debug("Get tensors for watch_point_hits.")
        tensors = []
        graph_name = None
        node_name = None
        watchpoint_ids = []
        for slot, tensor_hits in watchpoint_hits.items():
            if graph_name is None:
                graph_name = tensor_hits[0].graph_name
            if node_name is None:
                node_name = tensor_hits[0].node_name
            tensor_info = self._get_tensor_hit_info(slot, tensor_hits)
            tensors.append(tensor_info)
            for tensor_hit in tensor_hits:
                watch_point = tensor_hit.watchpoint
                watchpoint_id = watch_point['id']
                watchpoint_ids.append(watchpoint_id)

        watch_point_hits.append({
            'node_name': node_name,
            'tensors': tensors,
            'graph_name': graph_name,
            'graph_id': graph_name,
            'watchpoint_id': watchpoint_ids
        })

    @staticmethod
    def _get_tensor_hit_info(slot, tensor_hits):
        """
        Get watchpoint hit info of specified tensor.

        Args:
            slot (str): Slot id.
            tensor_hits (list): A list of watchpoint hit objects that the tensor hit.

        Returns:
            dict, tensor hit info.
        """
        res = {}
        watch_points = []

        for tensor_hit in tensor_hits:
            error_code = tensor_hit.error_code
            error_list = _get_error_list(error_code)
            watchpoint = tensor_hit.watchpoint
            watchpoint['error_code'] = error_code
            watchpoint['error_list'] = error_list
            watch_points.append(watchpoint)

        if watch_points:
            watch_points.sort(key=lambda watch_point: watch_point.get('id'))
            res = {
                'slot': slot,
                'watch_points': watch_points
            }
        return res

    def _is_tensor_hit(self, tensor_name, graph_name):
        """
        Check if the tensor is record in hit cache.

        Args:
            tensor_name (str): The name of ui tensor name.
            graph_name (str): The name of ui graph name

        Returns:
            bool, if the tensor is hit.
        """
        node_name, slot = tensor_name.rsplit(':', 1)
        index = self._multi_graph_hits.get((graph_name, node_name))
        if index is not None:
            watchpoint_hits = self._ordered_hits[index].get(slot)
            return bool(watchpoint_hits)
        return False

    def update_tensor_history(self, tensor_history):
        """
        Add hit flag to tensor history.

        Args:
            tensor_history (dict): The tensor history.
        """
        if not self._multi_graph_hits:
            return

        # add hit tensor names to `tensor_names`
        for tensor_info in tensor_history.get('tensor_history'):
            tensor_name = tensor_info['name']
            graph_name = tensor_info['graph_name']
            hit_flag = self._is_tensor_hit(tensor_name, graph_name)
            tensor_info['is_hit'] = hit_flag

    def get_tensor_hit_infos(self, tensor_name, graph_name):
        """
        Get all hit information of a tensor.

        Args:
            tensor_name (str): Tensor name showed on UI.

        Returns:
            dict, tensor hit info.
        """
        tensor_hit_info = {}
        if self._is_tensor_hit(tensor_name, graph_name):
            node_name, slot = tensor_name.rsplit(':', 1)
            tensor_hits = self._get_watchpoints_by_tensor_name(graph_name, node_name, slot)
            tensor_hit_info = self._get_tensor_hit_info(slot, tensor_hits)
        return tensor_hit_info


def validate_watch_condition(condition_mgr, watch_condition):
    """Validate watch condition."""
    if not isinstance(watch_condition, dict):
        log.error("<watch_condition> should be dict. %s received.", watch_condition)
        raise DebuggerParamTypeError("<watch_condition> should be dict.")
    # validate condition_id
    condition_id = watch_condition.get('id')
    if condition_id not in condition_mgr.conditions.keys():
        log.error("Invalid watch condition. Acceptable values are <%s>. %s received.",
                  str(condition_mgr.conditions.keys()), condition_id)
        raise DebuggerParamValueError("Invalid watch condition value.")
    # validate param
    validate_watch_condition_params(condition_mgr, watch_condition)


def validate_watch_condition_params(condition_mgr, watch_condition):
    """
    Validate watch condition parameters.

    Args:
        condition_mgr (ConditionMgr): Instance of ConditionMgr.
        watch_condition (dict): Watch condition.

            - id (str): Condition id. Should be in WATCHPOINT_CONDITION_MAPPING.
            - param (list): Condition value. Should be given for comparison condition. The value
                will be translated to np.float32.
    """
    condition_id = watch_condition.get('id')
    params = watch_condition.get('params')
    condition = condition_mgr.get_condition(condition_id)
    if condition_id in condition_mgr.get_no_param_condition():
        if params:
            log.error("No param is expected for %s condition", condition_id)
            raise DebuggerParamValueError("No param is expected.")
        return

    check_param_num = 0
    support_params = set()
    defined_support_params = set()
    range_param = {RANGE_START: None, RANGE_END: None}
    for param in params:
        if len(param) > 2:
            log.error("Invalid param keys for condition: %s", condition_id)
            raise DebuggerParamValueError("Invalid param keys.")

        condition_param_name = param.get("name")
        if condition_param_name not in condition.names:
            log.error("Invalid name of parameter for condition: %s, available values: %s",
                      condition_id, condition.names)
            raise DebuggerParamValueError("Invalid name of parameter.")

        condition_param = condition.get_parameter_definition(condition_param_name)
        validate_param_type(condition_id, condition_param, param)

        if not condition_param.is_valid(param.get("value")):
            log.error("Param %s out of range for condition: %s", condition_param_name, condition_id)
            raise DebuggerParamValueError("Parameter out of range.")

        if condition_param.param_type == ParamTypeEnum.CHECK_PARAM.value:
            if condition_param.required_params:
                defined_support_params = set(condition_param.required_params)
            check_param_num += 1
        else:
            support_params.add(condition_param.name)

        if condition_param_name in range_param:
            range_param[condition_param_name] = param.get("value")

        if check_param_num > 1:
            log.error("Multiple check params for condition: %s", condition_id)
            raise DebuggerParamValueError("Multiple check params.")

    if support_params != defined_support_params:
        log.error("Invalid support params for condition: %s", condition_id)
        raise DebuggerParamValueError("Invalid support params.")

    if range_param.get(RANGE_START) is not None and \
            range_param.get(RANGE_END) is not None and range_param.get(RANGE_START) > \
            range_param.get(RANGE_END):
        log.error("Invalid support params for condition: %s", condition_id)
        raise DebuggerParamValueError("Invalid support params.")


def validate_param_type(condition_id, condition_param, param):
    """
    Validate parameter type.

    Args:
        condition_id (str): Condition id. Should be in WATCHPOINT_CONDITION_MAPPING.
        condition_param (ConditionParameter): Condition Parameter object.
        param (dict): Condition parameter value.
    """
    if condition_param.type.name in (ValueTypeEnum.FLOAT64.name, ValueTypeEnum.INT64.name) \
            and (not isinstance(param.get("value"), (float, int)) or isinstance(param.get("value"), bool)):
        log.error("Number param should be given for condition: %s", condition_id)
        raise DebuggerParamValueError("Number param should be given.")
    if condition_param.type.name == ValueTypeEnum.BOOL.name \
            and not isinstance(param.get("value"), bool):
        log.error("Bool param should be given for condition: %s", condition_id)
        raise DebuggerParamValueError("Bool param should be given.")


def set_default_param(condition_mgr, watch_condition):
    """
    Set default param.
    Args:
        condition_mgr (ConditionMgr): Instance of ConditionMgr.
        watch_condition (dict): The watch condition.
            "condition": {
                id: "tensor_too_large",
                "params": [
                    {
                        "name": "abs_mean_gt",
                        "value": 1.1
                    }
                ]
            }

            - id (str): Id of condition.
            - param (list[dict]): The list of param for this condition.

    Returns:
        dict, the new watch_condition.
    """
    condition_id = watch_condition.get('id')
    condition = condition_mgr.get_condition(condition_id)
    for param in condition.parameters:
        if not param.visible_on_ui and not param.support_disable:
            watch_condition["params"].append({
                "name": param.name,
                "value": param.default_value
            })
    watch_condition["abbr"] = condition.abbr
    return watch_condition


def _get_error_list(error_code):
    """
    Get error list.
    Args:
        error_code (int): The code of errors.

    Returns:
        list, the error list.
    """
    all_error_list = ["nan", "inf", "no_prev_tensor", "out_of_mem", "no_history"]
    error_list = []
    for i, error_str in enumerate(all_error_list):
        error = (error_code >> i) & 1
        if error == 1:
            error_list.append(error_str)

    return error_list


def validate_rank_id(rank_id, device_amount):
    """validate rank id"""
    if rank_id >= device_amount:
        log.debug("The rank id %d over device amount.", rank_id)
